# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import asyncio
import functools

import urwid

from ..settings.defaults import DEFAULT_TAB_COMMANDS

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


def run(command_runner):
    """
    Run commands and start TUI

    Return False if any of the commands failed, True otherwise.
    """
    from .. import objects
    from . import tuiobjects

    # Don't catch theme.ParserError - a broken default theme should make us
    # croak obviously and horribly
    tuiobjects.theme.init(objects.localcfg.default('tui.theme'), tuiobjects.urwidscreen)

    # Load tui-specific hooks before commands run (commands may trigger hooks)
    from . import hooks  # noqa

    try:
        if not command_runner():
            return False
    # Make 'quit' behave as expected
    except urwid.ExitMainLoop:
        return True

    # Start logging to TUI widget instead of stdout/stderr
    tuiobjects.logwidget.enable()

    tuiobjects.topbar.help.update()

    # If no tab has been opened by cli or rc file, open default tabs
    if len(tuiobjects.tabs) <= 0:
        for cmd in DEFAULT_TAB_COMMANDS:
            objects.cmdmgr.run_sync(cmd, on_error=log.error)

    try:
        # Start polling torrent lists, counters, bandwidth usage, etc.
        asyncio.get_event_loop().run_until_complete(objects.srvapi.start_polling())
        old = tuiobjects.urwidscreen.tty_signal_keys('undefined','undefined',
                                                     'undefined','undefined','undefined')
        tuiobjects.urwidloop.run()
    finally:
        tuiobjects.urwidscreen.tty_signal_keys(*old)
        tuiobjects.logwidget.disable()

    return True


def redraw_screen(func):
    """
    Decorator for functions/methods that change screen contents

    https://github.com/urwid/urwid/pull/541
    https://github.com/urwid/urwid/pull/418
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        finally:
            asyncio.get_event_loop().call_soon(_redraw_screen)

    return wrapper

def _redraw_screen():
    try:
        from .tuiobjects import urwidloop
        urwidloop.draw_screen()
    except (AssertionError, RuntimeError):
        # catches trying to redraw the screen before urwidloop has started
        pass



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import functools
import os

from ... import objects
from ...completion import candidates
from ...utils.cliparser import Arg
from ..base import torrent as base
from . import _mixin as mixin
from ._common import make_tab_title_widget

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


class AddTorrentsCmd(base.AddTorrentsCmdbase,
                     mixin.polling_frenzy, mixin.make_request):
    provides = {'tui'}

    @staticmethod
    def make_path_absolute(path):
        # In the TUI, it makes more sense to use $HOME as the base directory for
        # relative paths instead of the current working directory.
        abspath = os.path.join(os.environ.get('HOME', '.'),
                               os.path.normpath(os.path.expanduser(path)))
        if os.path.exists(abspath):
            return abspath
        else:
            return path

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        return candidates.fs_path(args.curarg.before_cursor,
                                  base=os.environ.get('HOME', '.'),
                                  glob=r'*.torrent')


class TorrentDetailsCmd(base.TorrentDetailsCmdbase,
                        mixin.select_torrents, mixin.make_request):
    provides = {'tui'}

    async def display_details(self, torrent_id):
        make_titlew = functools.partial(make_tab_title_widget,
                                        attr_unfocused='tabs.torrentdetails.unfocused',
                                        attr_focused='tabs.torrentdetails.focused')

        from ...tui.tuiobjects import keymap, tabs
        from ...tui.views import TorrentDetailsWidget
        TorrentDetailsWidget_keymapped = keymap.wrap(TorrentDetailsWidget,
                                                     context='torrent')
        title_str = self.title if hasattr(self, 'title') else None
        detailsw = TorrentDetailsWidget_keymapped(torrent_id, title=title_str)
        tabid = tabs.load(make_titlew(detailsw.title), detailsw)
        tabs.set_info(command=self.command)

        def set_tab_title(text):
            # set_title() throws IndexError if the tab was removed, which may
            # have happened while TorrentDetailsWidget was waiting for a
            # response.
            try:
                tabs.set_title(make_titlew(text), position=tabid)
            except IndexError:
                pass
        detailsw.title_updater = set_tab_title


class ListTorrentsCmd(base.ListTorrentsCmdbase,
                      mixin.select_torrents,
                      mixin.create_list_widget):
    provides = {'tui'}

    def make_torrent_list(self, tfilter, sort, columns):
        from ...tui.views import TorrentListWidget
        self.create_list_widget(TorrentListWidget, theme_name='torrentlist',
                                tfilter=tfilter, sort=sort, columns=columns,
                                markable_items=True)


class TorrentMagnetURICmd(base.TorrentMagnetURICmdbase,
                          mixin.select_torrents):
    provides = {'tui'}

    def display_uris(self, uris):
        for uri in uris:
            self.info(uri)


class MoveTorrentsCmd(base.MoveTorrentsCmdbase,
                      mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}

    @classmethod
    async def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        def dest_path_candidates(curarg):
            return candidates.fs_path(curarg.before_cursor,
                                      base=objects.cfg['srv.path.complete'],
                                      directories_only=True,
                                      expand_home_directory=False)

        curarg = args.curarg
        if len(args) >= 3:
            if args.curarg_index == 1:
                return await candidates.torrent_filter(curarg)
            elif args.curarg_index == 2:
                return dest_path_candidates(curarg)
        elif len(args) == 2:
            # Single argument may be a path or a filter
            filter_cands = await candidates.torrent_filter(curarg)
            path_cands = dest_path_candidates(curarg)
            return (path_cands,) + filter_cands


class RemoveTorrentsCmd(base.RemoveTorrentsCmdbase,
                        mixin.polling_frenzy, mixin.make_request, mixin.select_torrents,
                        mixin.ask_yes_no):
    provides = {'tui'}
    CONFIRMATION_TAB_TITLE = 'Removal Confirmation'

    async def show_list_of_hits(self, tfilter):
        cmd = 'tab --title %r ls --sort name %s' % (self.CONFIRMATION_TAB_TITLE, tfilter)
        await objects.cmdmgr.run_async(cmd)

    async def remove_list_of_hits(self):
        cmd = 'tab --close %r --focus left' % self.CONFIRMATION_TAB_TITLE
        await objects.cmdmgr.run_async(cmd)


class RenameCmd(base.RenameCmdbase,
                mixin.polling_frenzy, mixin.make_request, mixin.select_torrents, mixin.select_files):
    provides = {'tui'}

    @classmethod
    async def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        # We don't care about options
        args = args.posargs()

        # If there is only one argument and it doesn't contain a path separator,
        # that means the user might want to rename the focused torrent, file or
        # directory.  In that case, the first argument is the destination and
        # the source is picked from the TUI.
        if len(args) == 2 and args.curarg_index == 1 and '/' not in args.curarg:
            source = cls._get_focused_item_source()
            if source is not None:
                log.debug('Getting destination candidates from TUI: %r', source)
                source_arg = Arg(source, curpos=len(source))
                tui_cands = await candidates.torrent_path(source_arg, only='auto')
                regular_cands = await candidates.torrent_path(args.curarg, only='any')
                return (tui_cands, regular_cands)

        # Any other arguments are handled identically in CLI and TUI
        log.debug('Not generating TUI-specific candidates')
        return await super().completion_candidates_posargs(args)

    @classmethod
    def _get_focused_item_source(cls):
        item_type = cls.get_focused_item_type()
        if item_type == 'torrent':
            data = cls.get_focused_item_data()
            source = 'id=%d' % (data['id'],)
            log.debug('  (focused torrent: %r)', source)
            return source

        elif item_type in ('file', 'directory'):
            data = cls.get_focused_item_data()
            # If there is no path separator in 'path-relative', that means focus
            # is on the torrent's name (or top-level directory)
            if '/' in data['path-relative']:
                focused_path = '/'.join(data['path-relative'].split('/')[1:])
                source = 'id=%d/%s' % (data['tid'], focused_path)
                log.debug('  (focused file/directory: %r)', source)
                return source


class StartTorrentsCmd(base.StartTorrentsCmdbase,
                       mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class StopTorrentsCmd(base.StopTorrentsCmdbase,
                      mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class VerifyTorrentsCmd(base.VerifyTorrentsCmdbase,
                        mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class LabelCmd(base.LabelCmd, mixin.make_request, mixin.select_torrents, mixin.polling_frenzy):
    provides = {'tui'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import asyncio
from collections import abc, defaultdict
from types import SimpleNamespace
from urllib.parse import quote_plus as urlquote

import blinker

from . import ClientError, utils

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


class TorrentBase(abc.Mapping):
    """Information about a torrent as a mapping

    This is the base class that all API implementations should use.

    Derivatives of this base class must add the methods 'update',
    '__getitem__' and '__iter__'.
    """

    TYPES = {
        'id'                           : int,
        'hash'                         : utils.SHA1,
        'name'                         : utils.SmartCmpStr,
        'ratio'                        : utils.Ratio,
        'status'                       : utils.Status,
        'path'                         : utils.SmartCmpPath,
        'private'                      : bool,
        'comment'                      : utils.SmartCmpStr,
        'creator'                      : utils.SmartCmpStr,
        'magnetlink'                   : str,
        'count-pieces'                 : utils.Int,

        '%downloaded'                  : utils.Percent,
        '%uploaded'                    : utils.Percent,
        '%metadata'                    : utils.Percent,
        '%verified'                    : utils.Percent,
        '%available'                   : utils.Percent,

        'peers-connected'              : utils.Int,
        'peers-uploading'              : utils.Int,
        'peers-downloading'            : utils.Int,
        'peers-seeding'                : utils.Count,

        'timespan-eta'                 : utils.Timedelta,
        'timespan-seeding'             : utils.Timedelta,
        'timespan-downloading'         : utils.Timedelta,
        'time-created'                 : utils.Timestamp,
        'time-added'                   : utils.Timestamp,
        'time-started'                 : utils.Timestamp,
        'time-activity'                : utils.Timestamp,
        'time-completed'               : utils.Timestamp,
        'time-manual-announce-allowed' : utils.Timestamp,

        'rate-down'                    : utils.BandwidthInBytes,
        'rate-up'                      : utils.BandwidthInBytes,
        'limit-rate-down'              : utils.BandwidthInBytesOrNone,
        'limit-rate-up'                : utils.BandwidthInBytesOrNone,
        'size-final'                   : utils.SizeInBytes,
        'size-total'                   : utils.SizeInBytes,
        'size-downloaded'              : utils.SizeInBytes,
        'size-uploaded'                : utils.SizeInBytes,
        'size-available'               : utils.SizeInBytes,
        'size-left'                    : utils.SizeInBytes,
        'size-corrupt'                 : utils.SizeInBytes,
        'size-piece'                   : utils.SizeInBytes,

        'error'                        : str,
        'trackers'                     : tuple,
        'peers'                        : tuple,
        'files'                        : None,

        'labels'                       : set,
    }

    def update(self, raw_torrent):
        raise NotImplementedError()

    def __getitem__(self, key):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()

    def __repr__(self):
        r = '<{} #{}'.format(type(self).__name__, self['id'])
        if 'name' in self:
            r += ' ' + repr(self['name'])
        return r + '>'

    def __eq__(self, other):
        if hasattr(other, 'get'):
            return self['id'] == other.get('id')
        return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if hasattr(other, 'get'):
            return self['id'] > other.get('id')
        return NotImplemented

    def __len__(self):
        return len(tuple(iter(self)))

    def __hash__(self):
        return hash(self['id'])


class TorrentAPIBase():
    async def get_magnet_uris(self, tfilter, tracker=False, trackers=True, name=True, size=True):
        response = await self.torrents(tfilter, keys=('hash', 'name', 'size-total', 'trackers'))
        if not response.success:
            raise ClientError(response.errors[0])
        else:
            torrents = response.torrents

        uris = []
        for t in torrents:
            parts = ['xt=urn:btih:%s' % t['hash']]
            if name and t['name'] != t['hash']:
                parts.append('dn=%s' % urlquote(t['name']))
            if size and t['size-total'] > 0:
                parts.append('xl=%r' % t['size-total'])
            if t['trackers']:
                if tracker:
                    # Including only one tracker
                    parts.append('tr=%s' % urlquote(str(t['trackers'][0]['url-announce'])))
                elif trackers:
                    # Including all trackers
                    for tracker in t['trackers']:
                        parts.append('tr=%s' % urlquote(str(tracker['url-announce'])))
            uris.append('magnet:?' + '&'.join(parts))
        return uris


class TorrentFileTreeBase(abc.Mapping):
    """Nested mapping of a Torrent's files"""

    # Distinguish subtrees from files without comparing classes everywhere
    # ("parent" or "leaf")
    nodetype = 'parent'

    def __init__(self, location, path, *args, **kwargs):
        self._location = location  # Absolute download location
        self._path = path          # Path relative to location
        self._items = NotImplemented

    @property
    def files(self):
        """Yield all TorrentFiles recursively"""
        for entry in self._items.values():
            if entry.nodetype == 'leaf':
                yield entry
            else:
                yield from entry.files

    @property
    def directories(self):
        """Yield (name, TorrentFileTree) tuples recursively"""
        for name,entry in self._items.items():
            if entry.nodetype == 'parent':
                yield (name, entry)
                yield from entry.directories

    @property
    def location(self):
        """Absolute path of the torrent; base directory"""
        return self._location

    @property
    def path(self):
        """Relative path in the torrent"""
        return self._path

    @property
    def id(self):
        return tuple(f['id'] for f in self.files)

    def __repr__(self):
        return '<%s path=%r: %r>' % (type(self).__name__, self._path, self._items)

    def __getitem__(self, key):
        return self._items[key]

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)


class FreeSpaceAPIBase():
    """
    Gather available storage space for multiple directories

    path_getters: Iterable of callables that must return a path (str) or None
    rpc: Object that can be used by the implementation of get_free_space()
    settings: Object with an on_update() method that allows us to register a callback for
              path changes
    """
    def __init__(self, path_getters, rpc, settings):
        self._path_getters = tuple(path_getters)
        self._rpc = rpc
        self._on_update = blinker.Signal()
        self._info = defaultdict(lambda: SimpleNamespace(path=None, free=None, error=None))
        settings.on_update(self._gather_info_wrapper)

    def _gather_info_wrapper(self, settings):
        asyncio.ensure_future(self._gather_info_wrapper_coro())

    async def _gather_info_wrapper_coro(self):
        infos = {}
        for path_getter in self._path_getters:
            try:
                path = path_getter()
            except ClientError as e:
                log.debug('Exception from %r: %r', path_getter, e)
            else:
                if not path:
                    log.debug('Ignoring false path: %r: %r', path, bool(path))
                elif not isinstance(path, str):
                    log.debug('Ignoring invalid path: %r: %r', type(path), path)
                else:
                    try:
                        free = await self.get_free_space(path)
                    except ClientError as e:
                        infos[path] = SimpleNamespace(path=path, free=None, error=e)
                    else:
                        infos[path] = SimpleNamespace(path=path,
                                                      free=utils.convert.size(free, unit='byte'),
                                                      error=None)
        if self._info != infos:
            self._info = infos
            self._on_update.send(self)

    @property
    def info(self):
        """
        Dictionary that maps paths to namespaces

        Each namespace has these attributes:

          path:  Directory returned by one of the path_getters or None if the path_getter
                 failed
          free:  Available space in path or None if get_free_space() or the path_getter
                 failed
          error: Exception object raised by the path_getter or get_free_space()
        """
        return self._info

    async def get_free_space(self, path):
        """
        Return the available space in directory `path` in bytes
        """
        raise NotImplementedError()

    def on_update(self, callback, autoremove=True):
        """
        Register `callback` to be called when free space has changed

        `callback` gets the instance of this class.

        If `autoremove` is True, `callback` is removed automatically when it is deleted.
        """
        log.debug('Registering %r to receive free space updates', callback)
        self._on_update.connect(callback, weak=autoremove)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""
Create and manage commands

Commands can be created anywhere by defining a class with metaclass set to
CommandMeta.

>>> class MyCommand(metaclass=CommandMeta):
>>>     ...

Every command class must have a 'run' method, which is called when the command
is executed. It can be a coroutine function or a normal, synchronous function.
The command's arguments are passed to 'run' as keywords.

The return value of the 'run' method is ignored.  It must raise CmdError on
failure.  The argument to CmdError is used as the error message.  It may also
print errors via its 'error' method or informational messages via its 'info'
method.

If a command executes multiple actions (e.g. pausing two torrents) and any of
those actions fails, it should be a failure (e.g. first torrent was paused,
second torrent does not exist, third torrent was paused -> return False because
of second torrent).

Every command class must have the following class attributes:

    run (callable):            Method that is called when the command runs
    name (string):             Name of the command
    category (string):         Name of the command category (used by `help commands`)
    provides (set of strings): Supported interfaces ('cli' and/or 'tui')
    description (string):      One-line description for command

These class attributes are optional:

    aliases (sequence of strings):   Alternative command names
    argspecs (sequence of mappings): argparse specs for arguments
    usage (sequence of strings):     Syntax string of any arguments (in man pages
                                     this is called SYNOPSIS)
    examples (sequence of strings):  Typical invocations
    more_sections (dict):            Maps additional section names to list of strings or
                                     callables that return lists of strings

Arguments
---------

Each command class gets its own argparser.ArgumentParser instance.  Command
arguments are specified by setting the 'argspecs' class attribute to a
sequence of mappings.  The mappings are then used to create arguments for
ArgumentParser.add_argument.  Most items are passed on unmodified, but there
are exceptions:

- 'names' must be set to a string or a sequence of strings that is passed on
  as the positional arguments to add_argument.  The first string is used as
  the keyword when providing the argument to the run method.

- 'description' must be set to a string that describes what the argument does.

- 'nargs' can be the string 'REMAINDER', which is replaced with the value of
  argparse.REMAINDER.
"""

# flake8: noqa

OPS_AND = ('&', 'and')
OPS_OR  = ('|', 'or')
OPS_SEQ = (';', 'also')
OPS = OPS_AND + OPS_OR + OPS_SEQ

from .cmdbase import CommandMeta, _CommandBase
from .cmderror import *
from .cmdmanager import CommandManager
from .utils import is_cmdcls, is_op



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import concurrent
import socket

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)

_cache = {}
_lookup_pool = concurrent.futures.ThreadPoolExecutor(max_workers=10)

def gethostbyaddr(ip):
    hostname = _cache.get(ip)
    if hostname is None:
        try:
            hostname = socket.gethostbyaddr(ip)[0]
        except OSError:
            hostname = ip
        finally:
            # I've seen IPs being resolved to ".", dunno why.
            if hostname == '.':
                hostname = ip
            _cache[ip] = hostname
    return hostname

def gethostbyaddr_from_cache(ip):
    return _cache.get(ip)

def query(*ips, callback=None):
    def cb(fut):
        hostname = fut.result()
        if callback:
            callback(hostname)

    for ip in ips:
        fut = _lookup_pool.submit(gethostbyaddr, ip)
        fut.add_done_callback(cb)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""TUI and CLI specs for file list columns"""

import os

from . import ColumnBase, _ensure_hide_unit

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


COLUMNS = {}
ALIASES = {'n'    : 'name', 'filename': 'name',
           'sz'   : 'size',
           'dn'   : 'downloaded',
           '%dn'  : '%downloaded',
           'prio' : 'priority',
           'mark' : 'marked'}


class Filename(ColumnBase):
    header = {'left': 'Filename'}
    align = 'left'
    width = None
    min_width = 10
    may_have_wide_chars = True

    def get_value(self):
        return self.data['name']

COLUMNS['name'] = Filename


class Size(ColumnBase):
    header = {'left': 'Size', 'right': '?'}
    width = 6
    min_width = 6

    def get_value(self):
        return self._from_cache(_ensure_hide_unit, self.data['size-total'])

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['size'] = Size


class Downloaded(ColumnBase):
    header = {'left': 'Dn', 'right': '?'}
    width = 6
    min_width = 6

    def get_value(self):
        return self._from_cache(_ensure_hide_unit, self.data['size-downloaded'])

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['downloaded'] = Downloaded


class PercentDownloaded(ColumnBase):
    header = {'right': '%'}
    width = 4
    min_width = 4

    def get_value(self):
        return self._from_cache(_ensure_hide_unit, self.data['%downloaded'])

COLUMNS['%downloaded'] = PercentDownloaded


class Priority(ColumnBase):
    header = {'left': 'Priority'}
    width = 4
    min_width = 4
    align = 'left'

    def get_value(self):
        val = self.get_raw_value()
        return '' if val == 'normal' else val

    def get_raw_value(self):
        return 'off' if self.data['is-wanted'] is False else self.data['priority']

COLUMNS['priority'] = Priority


class Marked(ColumnBase):
    interfaces = ('tui',)

COLUMNS['marked'] = Marked


class TorrentFileDirectory(dict):
    """
    A mapping with the same keys as a TorrentFile instance but represents a directory

    Values recursively summarize the values of all the TorrentFiles in the subtree.
    """
    nodetype = 'parent'

    def __init__(self, name, tree, filtered_count=0):
        tfiles = tuple(tree.files)
        self.update({
            'id'              : tree.id,
            'tid'             : tfiles[0]['tid'],
            'name'            : self.create_directory_name(name, filtered_count),
            'path-absolute'   : os.path.join(tree.location, tree.path),
            'path-relative'   : tree.path,
            'location'        : tree.location,
            'size-total'      : self._sum_size(tfiles, 'size-total'),
            'size-downloaded' : self._sum_size(tfiles, 'size-downloaded'),
            'is-wanted'       : True,
            'priority'        : self._sum_priority(tfiles),
        })
        perc_dl_cls = type(tfiles[0]['%downloaded'])
        try:
            self['%downloaded'] = perc_dl_cls(self['size-downloaded'] / self['size-total'] * 100)
        except ZeroDivisionError:
            self['%downloaded'] = perc_dl_cls(0)

    @staticmethod
    def _sum_size(tfiles, key):
        sizes = tuple(tfile[key] for tfile in tfiles)
        # Preserve the original type (Float)
        first_size = sizes[0]
        start_value = type(first_size)(0, unit=first_size.unit, prefix=first_size.prefix)
        return sum(sizes, start_value)

    @staticmethod
    def _sum_priority(tfiles):
        if len(set(tfile['priority'] for tfile in tfiles)) == 1:
            return tfiles[0]['priority']
        else:
            return ''

    @staticmethod
    def create_directory_name(name, filtered_count):
        if filtered_count > 0:
            return '%s (%d file%s filtered)' % (name, filtered_count,
                                                '' if filtered_count == 1 else 's')
        else:
            return str(name)

    def __hash__(self):
        return hash(self.get('path', ''))

    def __repr__(self):
        return '<%s %s>' % (type(self).__name__, self['path-absolute'])



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering PeerList items by various values"""

from .. import rdns
from ..ttypes import TorrentPeer
from .base import BoolFilterSpec, CmpFilterSpec, Filter, FilterChain, FilterSpecDict


def _cmp_host_or_ip(p, op, v):
    hostname = rdns.gethostbyaddr_from_cache(p['ip'])
    if hostname is None:
        return op(p['ip'], v)
    else:
        return op(hostname, v)

class _BoolFilterSpec(BoolFilterSpec):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, needed_keys=('peers',), **kwargs)

class _CmpFilterSpec(CmpFilterSpec):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, needed_keys=('peers',), **kwargs)

class _SingleFilter(Filter):
    DEFAULT_FILTER = 'host'

    BOOLEAN_FILTERS = FilterSpecDict({
        'all'         : _BoolFilterSpec(None,
                                        aliases=('*',),
                                        description='All peers'),
        'uploading'   : _BoolFilterSpec(lambda p: p['rate-up'] > 0,
                                        aliases=('upg',),
                                        description='Peers we are uploading to'),
        'downloading' : _BoolFilterSpec(lambda p: p['rate-down'] > 0,
                                        aliases=('dng',),
                                        description='Peers we are downloading from'),
        'seeding'     : _BoolFilterSpec(lambda p: p['%downloaded'] >= 100,
                                        aliases=('sdg',),
                                        description='Peers that have downloaded all data'),
    })

    COMPARATIVE_FILTERS = FilterSpecDict({
        'downloaded'  : _CmpFilterSpec(value_getter=lambda p: p['downloaded'],
                                       value_type=TorrentPeer.TYPES['downloaded'],
                                       as_bool=lambda p: p['downloaded'] > 0,
                                       aliases=('dn',),
                                       description='Match VALUE against number of bytes peer has downloaded'),
        '%downloaded' : _CmpFilterSpec(value_getter=lambda p: p['%downloaded'],
                                       value_type=TorrentPeer.TYPES['%downloaded'],
                                       aliases=('%dn',),
                                       description='Match VALUE against percentage of bytes peer has downloaded'),
        'client'      : _CmpFilterSpec(value_getter=lambda p: p['client'],
                                       value_type=TorrentPeer.TYPES['client'],
                                       aliases=('cl',),
                                       description='Match VALUE against peer client'),
        'host'        : _CmpFilterSpec(value_getter=lambda p: rdns.gethostbyaddr_from_cache(p['ip']) or p['ip'],
                                       value_type=TorrentPeer.TYPES['ip'],
                                       description='Match VALUE against peer host name or IP address'),
        'port'        : _CmpFilterSpec(value_getter=lambda p: p['port'],
                                       value_type=TorrentPeer.TYPES['port'],
                                       description='Match VALUE against peer port'),
    })


class PeerFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = _SingleFilter



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import asyncio

from ... import objects
from ...completion import candidates
from .. import CmdError, CommandMeta
from . import _mixin as mixin
from ._common import make_COLUMNS_doc, make_SCRIPTING_doc, make_X_FILTER_spec

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


class ListFilesCmdbase(mixin.get_file_columns, metaclass=CommandMeta):
    name = 'filelist'
    aliases = ('fls', 'lsf')
    provides = set()
    category = 'file'
    description = 'List files of torrent(s)'
    usage = ('filelist [<OPTIONS>]',
             'filelist [<OPTIONS>] <TORRENT FILTER>',
             'filelist [<OPTIONS>] <TORRENT FILTER> <FILE FILTER>')
    examples = ('filelist',
                'filelist size<100MB',
                'filelist A.Torrent.with.Files priority=low')
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='?'),
        make_X_FILTER_spec('FILE', or_focused=False, nargs='?'),
        {'names': ('--columns', '-c'),
         'default_description': "current value of 'columns.files' setting",
         'description': ('Comma-separated list of column names '
                         "(see COLUMNS section)")},
    )

    from ...views.file import COLUMNS
    more_sections = {
        'COLUMNS': make_COLUMNS_doc(COLUMNS, '--columns', 'columns.files'),
        'SCRIPTING': make_SCRIPTING_doc(name),
    }

    async def run(self, TORRENT_FILTER, FILE_FILTER, columns):
        columns = objects.localcfg['columns.files'] if columns is None else columns
        try:
            columns = self.get_file_columns(columns)
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
            ffilter = self.select_files(FILE_FILTER,
                                        allow_no_filter=True,
                                        discover_file=False)
        except ValueError as e:
            raise CmdError(e)

        log.debug('Listing %s files of %s torrents', ffilter, tfilter)

        if asyncio.iscoroutinefunction(self.make_file_list):
            await self.make_file_list(tfilter, ffilter, columns)
        else:
            self.make_file_list(tfilter, ffilter, columns)

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        posargs = args.posargs({('--columns', '-c'): 1})
        if posargs.curarg_index == 1:
            return candidates.torrent_filter(args.curarg)
        elif posargs.curarg_index == 2:
            torrent_filter = posargs[1]
            return candidates.file_filter(args.curarg, torrent_filter)

    @classmethod
    def completion_candidates_params(cls, option, args):
        """Complete parameters (e.g. --option parameter1,parameter2)"""
        if option == '--columns':
            return candidates.column_names('files')


class PriorityCmdbase(metaclass=CommandMeta):
    name = 'priority'
    aliases = ('prio',)
    provides = set()
    category = 'file'
    description = 'Change download priority of files'
    usage = ('priority <PRIORITY>',
             'priority <PRIORITY> <TORRENT FILTER>',
             'priority <PRIORITY> <TORRENT FILTER> <FILE FILTER>')
    examples = ('priority low',
                'priority high "that torrent" size>12M')
    _PRIORITY = {'off'    : ('o', '0', 'off'),
                 'low'    : ('l', '-', 'low'),
                 'normal' : ('n', '=', 'normal'),
                 'high'   : ('h', '+', 'high')}
    argspecs = (
        {'names': ('PRIORITY',),
         'description': 'File priority; must be %s, %s or %s' % (
             '/'.join(p[2] for p in _PRIORITY.values()),
             '/'.join(p[0] for p in _PRIORITY.values()),
             '/'.join(p[1] for p in _PRIORITY.values()))},
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='?'),
        make_X_FILTER_spec('FILE', or_focused=True, nargs='?'),
    )

    async def run(self, PRIORITY, TORRENT_FILTER, FILE_FILTER):
        priority = None
        for p,names in self._PRIORITY.items():
            if PRIORITY in names:
                priority = p
                break
        if priority is None:
            raise CmdError('Invalid priority: %r' % PRIORITY)

        # Whether the user manually typed a filter
        utilize_tui = not bool(TORRENT_FILTER)

        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)

            # If the user specified a filter instead of selecting via the TUI,
            # ignore focused/marked files.
            log.debug('%sdiscovering file(s)', '' if utilize_tui else 'Not ')
            ffilter = self.select_files(FILE_FILTER,
                                        allow_no_filter=True,
                                        discover_file=utilize_tui)
        except ValueError as e:
            raise CmdError(e)

        if not utilize_tui:
            self.info('New download priority of %s files in %s torrents: %s' %
                      ('all' if ffilter is None else ffilter, tfilter, priority))
            quiet = False
        else:
            # We're operating on focused or marked files and changes are
            # indicated by the updated file list, so no info messages
            # necessary.
            quiet = True

        log.debug('Setting file download priority to %s for %s files of %s torrents',
                  priority, ffilter, tfilter)
        response = await self.make_request(
            objects.srvapi.torrent.file_priority(tfilter, ffilter, priority),
            polling_frenzy=True, quiet=quiet)
        if not response.success:
            raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        if args.curarg_index == 1:
            # "off", "low", "normal" or "high"
            return candidates.Candidates((p[2] for p in cls._PRIORITY.values()),
                                         label='Priority')
        elif args.curarg_index == 2:
            return candidates.torrent_filter(args.curarg)
        elif args.curarg_index == 3:
            torrent_filter = args[2]
            return candidates.file_filter(args.curarg, torrent_filter)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from types import SimpleNamespace

from ._converter import DataSizeConverter

convert = SimpleNamespace(bandwidth=DataSizeConverter(),
                          size=DataSizeConverter())


def cached_property(fget=None, *, after_creation=None):
    """
    Property that replaces itself with the requested value when accessed

    `after_creation` is called with the instance of the property when the
    property is accessed for the first time.
    """
    # https://stackoverflow.com/a/6849299
    class _cached_property():
        def __init__(self, fget):
            self._fget = fget
            self._property_name = fget.__name__
            self._after_creation = after_creation
            self._cache = {}

        def __get__(self, obj, cls):
            value = self._fget(obj)
            setattr(obj, self._property_name, value)
            if self._after_creation is not None:
                self._after_creation(obj)
            return value

    if fget is None:
        return _cached_property
    else:
        return _cached_property(fget)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Commands that work exclusively in the TUI"""

import functools
import os
import shlex
from functools import partial

from ... import client, objects
from ...completion import candidates
from .. import CmdError, CommandMeta, utils
from . import _mixin as mixin
from ._common import make_tab_title_widget

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


# Import tui.main module only on demand
def _get_keymap_contexts():
    from ...tui.tuiobjects import keymap
    return tuple(keymap.contexts)


class BindCmd(metaclass=CommandMeta):
    name = 'bind'
    provides = {'tui'}
    category = 'tui'
    description = 'Bind keys to commands or other keys'
    usage = ('bind [<OPTIONS>] <KEY> <ACTION>',)
    examples = ('bind ctrl-a tab ls active',
                'bind --context tabs alt-[ tab --focus left',
                'bind --context tabs alt-] tab --focus right',
                'bind --context torrent alt-! start --force',
                "bind --context torrent 'd .' rm",
                "bind --context torrent 'd+!' rm --delete-files",
                'bind u <up>',
                'bind d <down>')
    argspecs = (
        {'names': ('--context','-c'),
         'description': 'Where KEY is grabbed (see CONTEXTS section)'},
        {'names': ('--description','-d'),
         'description': 'Explanation of what ACTION does'},
        {'names': ('KEY',),
         'description': 'One or more keys or key combinations (see KEYS section)'},
        {'names': ('ACTION',), 'nargs': 'REMAINDER',
         'description': ("Any command or '<KEY>' (including the brackets) "
                         'to translate one key to another')},
    )

    def __create_CONTEXTS_section():
        lines = [
            ('The same key can be bound multiple times in different contexts.  '
             'With no context given, the default context is used.  The default '
             "context gets the key if it isn't mapped in any other relevant context."),
            '',
            'Available contexts are: ' + ', '.join(str(c) for c in _get_keymap_contexts()),
            '',
            'EXAMPLE',
            '\tbind --context torrent ctrl-t start',
            '\tbind --context tabs ctrl-t tab',
            '\tbind ctrl-t <left>',
            '',
            ('\tWhen focusing a torrent, <ctrl-t> starts the focused torrent.  '
             'If focus is not on a torrent but still on a tab (e.g. in an empty '
             'torrent list or when reading this text) a new tab is opened.  '
             'Otherwise (e.g. focus is on the command prompt), <ctrl-t> does the '
             'same as <left>.'),
        ]
        return lines

    more_sections = {
        'CONTEXTS': __create_CONTEXTS_section,
        'KEYS': (
            'Single-character keys are specified as themselves (e.g. h, X, 5, !, þ, ¥, etc).',
            '',
            ('Special key names are enter, space, tab, backspace, insert, delete, home, end, '
             'up, down, left, right, pgup, pgdn and f1-12.'),
            '',
            ("The modifiers 'ctrl', 'alt' and 'shift' are separated with '-' from the key "
             "(e.g. alt-i, shift-delete, ctrl-a).  shift-x is identical to X."),
            '',
            ("Chained keys are sparated by single spaces (' ') or pluses ('+') and must be "
             "given as one argument per chain."),
        )
    }

    def run(self, context, description, KEY, ACTION):
        from ...tui.tuiobjects import keymap
        key = KEY
        if len(ACTION) == 1 and ACTION[0][0] == '<' and ACTION[0][-1] == '>':
            # ACTION is another key (e.g. 'j' -> 'down')
            action = keymap.mkkey(ACTION[0])
        else:
            action = ' '.join(shlex.quote(x) for x in ACTION)

        if context is None:
            context = keymap.DEFAULT_CONTEXT
        elif context not in _get_keymap_contexts():
            raise CmdError('Invalid context: %r' % (context,))

        try:
            keymap.bind(key, action, context=context, description=description)
        except ValueError as e:
            raise CmdError(e)

    _own_options = {('--context', '-c'): 1,
                    ('--description', '-d'): 1}

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        posargs = args.posargs(cls._own_options)
        if posargs.curarg_index == 2:
            # First positional argument is the key, second is the command's name
            return candidates.commands()
        else:
            # Any other positional arguments will be passed to subcmd
            subcmd = cls._get_subcmd(args)
            if subcmd:
                return candidates.for_args(subcmd)

    @classmethod
    def completion_candidates_opts(cls, args):
        """Return candidates for arguments that start with '-'"""
        subcmd = cls._get_subcmd(args)
        if subcmd:
            # Get completion candidates from subcmd's class
            return candidates.for_args(subcmd)
        else:
            # Parent class generates candidates for our own options
            return super().completion_candidates_opts(args)

    @classmethod
    def completion_candidates_params(cls, option, args):
        """Complete parameters (e.g. --option parameter1,parameter2)"""
        if option == '--context':
            return candidates.keybinding_contexts()

    @classmethod
    def _get_subcmd(cls, args):
        # posarg[0] is 'bind', posarg[1] is the key
        subcmd_start = args.nth_posarg_index(3, cls._own_options)
        # Subcmd is only relevant if the cursor is somewhere on it.
        # Otherwise, we're on our own arguments.
        if subcmd_start is not None and subcmd_start < args.curarg_index:
            return args[subcmd_start:]


class UnbindCmd(metaclass=CommandMeta):
    name = 'unbind'
    provides = {'tui'}
    category = 'tui'
    description = 'Unbind keys so pressing them has no effect'
    usage = ('unbind [<OPTIONS>] <KEY> <KEY> ...',)
    examples = ('unbind --context main ctrl-l',
                'unbind q')
    argspecs = (
        {'names': ('--context','-c'),
         'description': 'Where KEY is grabbed (see "bind" command)'},
        {'names': ('--all','-a'), 'action': 'store_true',
         'description': 'Remove all keybindings or only those in given context'},
        {'names': ('KEY',), 'nargs': 'REMAINDER',
         'description': 'Keys or key combinations (see "bind" command)'},
    )

    more_sections = {
        'COMPLETE UNBINDING': (
            ('For this command there is a special context called \'all\' that '
             'unbinds the key for every context.'),
            '',
            'Note that \'unbind --all\' is very different from \'unbind --context all\''
        )
    }

    def run(self, context, all, KEY):
        from ...tui.tuiobjects import keymap

        if context is not None and context not in _get_keymap_contexts():
            raise CmdError('Invalid context: %r' % (context,))

        if KEY:
            if context:
                success = self._unbind_keys(keys=KEY, context=context)
            elif all:
                success = self._unbind_keys(keys=KEY, context=keymap.ALL_CONTEXTS)
            else:
                success = self._unbind_keys(keys=KEY, context=keymap.DEFAULT_CONTEXT)
        else:
            success = self._unbind_all_keys(context=context)

        if not success:
            raise CmdError()

    def _unbind_keys(self, keys, context):
        from ...tui.tuiobjects import keymap
        success = True
        for key in keys:
            try:
                keymap.unbind(key, context=context)
            except ValueError as e:
                self.error(e)
                success = False
        return success

    def _unbind_all_keys(self, context):
        from ...tui.tuiobjects import keymap
        if context is None:
            keymap.clear()
        else:
            keymap.clear(context=context)
        return True

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        return candidates.keybinding_keys(args)

    @classmethod
    def completion_candidates_params(cls, option, args):
        """Complete parameters (e.g. --option parameter1,parameter2)"""
        if option == '--context':
            return candidates.keybinding_contexts()


class SetCommandCmd(mixin.placeholders, metaclass=CommandMeta):
    name = 'setcommand'
    aliases = ('setcmd',)
    provides = {'tui'}
    category = 'tui'
    description = 'Open the command line and insert a command'
    usage = ('setcommand [--trailing-space] <COMMAND> <ARGUMENT> <ARGUMENT> ...',)
    examples = (
        'setcommand --trailing-space tab ls',
        '\tAsk the user for a filter before opening a new torrent list.',
        '',
        'setcommand move {{location}}/',
        ('\tMove the focused torrent, using the path of the currently focused '
         'list item as a starting point.'),
        '',
        'setcommand move id={{id}} {{location}}/',
        ('\tSame as above, but make sure to move the correct torrent in case '
         'it is removed from the list while typing in the new path, e.g. if '
         'we\'re listing active torrents and the focused torrent stops being active.'),
    )
    argspecs = (
        {'names': ('COMMAND',), 'nargs': 'REMAINDER',
         'description': 'Command the can user edit before executing it (see PLACEHOLDERS)'},
        {'names': ('--trailing-space', '-s'), 'action': 'store_true',
         'description': 'Append a space at the end of COMMAND'},
    )
    more_sections = {
        'PLACEHOLDERS': mixin.placeholders.HELP,
    }

    async def run(self, COMMAND, trailing_space):
        log.debug('Unresolved command: %r', COMMAND)
        args = await self.parse_placeholders(*COMMAND)
        log.debug('Command with resolved placeholders: %r', args)

        if args:
            cmdstr = ' '.join(shlex.quote(str(arg)) for arg in args)
            if trailing_space:
                cmdstr += ' '
            from ...tui.tuiobjects import widgets
            widgets.show('cli')
            widgets.cli.base_widget.edit_text = cmdstr
            widgets.cli.base_widget.edit_pos = len(cmdstr)

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        posargs = args.posargs()
        if posargs.curarg_index == 1:
            # First positional argument is the subcmd's name
            return candidates.commands()
        else:
            # Any other positional arguments are part of subcmd
            subcmd = cls._get_subcmd(args)
            if subcmd:
                return candidates.for_args(subcmd)

    @classmethod
    def completion_candidates_opts(cls, args):
        """Return candidates for arguments that start with '-'"""
        subcmd = cls._get_subcmd(args)
        if subcmd:
            # Get completion candidates for subcmd
            return candidates.for_args(subcmd)
        else:
            # Parent class generates candidates for our own options
            return super().completion_candidates_opts(args)

    @staticmethod
    def _get_subcmd(args):
        # First posarg is 'setcommand'
        subcmd_start = args.nth_posarg_index(2)
        # Subcmd is only relevant if the cursor is somewhere on it.
        # Otherwise, we're on our own arguments.
        if subcmd_start is not None and subcmd_start < args.curarg_index:
            return args[subcmd_start:]


class InteractiveCmd(mixin.placeholders, metaclass=CommandMeta):
    name = 'interactive'
    provides = {'tui'}
    category = 'tui'
    description = 'Complete partial command with user input from a dialog'
    usage = ('interactive <COMMAND> [<OPTIONS>]',)
    examples = (
        'interactive "move \'[{location}/]\'"',
        '\tAsk for the destination directory when moving torrents.',
        '',
        'tab ls & interactive "limit \'[]\'" --per-change --on-cancel "tab --close --focus left"',
        ('\tOpen a new tab with all torrents and filter them as you type.  '
         'Keep the tab open if the user input field is accepted with <enter> '
         'or close the tab and focus the previous one if the dialog is aborted '
         'with <escape>.'),
        '',
        'tab ls stopped & interactive \'limit "[]"\' -p -a "mark --all & start" -x "tab --close --focus left"',
        ('\tSearch for stopped torrents only.  When accepted, the matching torrents '
         'are started.  The new tab is always closed, whether the dialog is '
         'accepted or not.'),
    )
    argspecs = (
        {'names': ('COMMAND',),
         'description': ('Any command with "[PREFILLED TEXT]" as marker for '
                         'user input field (see USER INPUT FIELDS) and '
                         '"{{NAM}}" as placeholder for values of the currently '
                         'focused list item (see PLACEHOLDERS)')},
        {'names': ('--per-change', '-p'), 'action': 'store_true',
         'description': 'Whether to run COMMAND every time the input is changed'},
        {'names': ('--on-accept', '-a'), 'metavar': 'ACCEPT COMMAND',
         'description': 'Command to run when the dialog is accepted (with <enter>)'},
        {'names': ('--on-cancel', '-c'), 'metavar': 'CANCEL COMMAND',
         'description': 'Command to run when the dialog is aborted (with <escape>)'},
        {'names': ('--on-close', '-x'), 'metavar': 'CLOSE COMMAND',
         'description': 'Command to run after the dialog is closed either way'},
        {'names': ('--ignore-errors', '-i'), 'action': 'store_true',
         'description': 'Whether to ignore errors from COMMAND'},
    )
    more_sections = {
        'COMMANDS': (('For each occurrence of "[]" in any command, the user is '
                      'prompted for input to insert at that point.  Any text between '
                      '"[" and "]" is used as the initial user input.  "[" can be '
                      'escaped with "\\" in which case the corresponding "]" is also '
                      'interpreted literally.'),
                     '',
                     ('COMMAND is called if the user presses <enter> or, if --per-change '
                      'is given, after any user input field is changed.'),
                     '',
                     ('COMMAND must contain at least one user input field.  Any of the '
                      'commands described below are called without user interaction if '
                      'they don\'t contain any user input fields.'),
                     '',
                     ('ACCEPT COMMAND is called after COMMAND if the user accepts the '
                      'dialog by pressing <enter>.'),
                     '',
                     'CANCEL COMMAND is called if the user aborts the COMMAND dialog.',
                     '',
                     ('CLOSE COMMAND is always called when the dialog is closed either '
                      'by accepting or by cancelling it.')),
        'PLACEHOLDERS': mixin.placeholders.HELP,
    }

    import re
    _input_regex = re.compile(r'(?<!\\)(\[.*?\])')

    async def run(self, COMMAND, per_change, on_accept, on_cancel, on_close, ignore_errors):
        cmd = await self._parse_cmd(COMMAND)
        accept_cmd = await self._parse_cmd(on_accept) if on_accept else None
        cancel_cmd = await self._parse_cmd(on_cancel) if on_cancel else None
        close_cmd = await self._parse_cmd(on_close) if on_close else None
        self._ignore_errors = ignore_errors

        if len(cmd) == 1:
            # There are no user input markers
            raise CmdError('No user input fields ("[]"): %s' % COMMAND)

        def close_cb():
            self._run_cmd_or_open_dialog(close_cmd)

        if per_change:
            def accept_cb():
                self._run_cmd_from_dialog()
                self._run_cmd_or_open_dialog(accept_cmd)

            def cancel_cb():
                self._run_cmd_or_open_dialog(cancel_cmd)

            self._open_dialog(cmd,
                              on_change=self._run_cmd_from_dialog,
                              on_accept=accept_cb,
                              on_cancel=cancel_cb,
                              on_close=close_cb)
        else:
            def accept_cb():
                self._run_cmd_from_dialog()
                self._run_cmd_or_open_dialog(accept_cmd)

            def cancel_cb():
                self._run_cmd_or_open_dialog(cancel_cmd)

            self._open_dialog(cmd,
                              on_accept=accept_cb,
                              on_cancel=cancel_cb,
                              on_close=close_cb)

    _WIDGET_NAME = 'interactive_prompt'
    _MIN_EDIT_WIDTH = 25
    _MAX_EDIT_WIDTH = 50

    def _open_dialog(self, cmd, on_change=None, on_accept=None, on_cancel=None, on_close=None):
        import urwid

        from ...tui.cli import CLIEditWidget

        def accept_cb(widget):
            # CLIEditWidget only automatically appends to history when it gets
            # an <enter> key, but only one gets it if there are multiple user
            # input fields.
            for part in self._edit_widgets:
                part.append_to_history()
            self._close_dialog()
            if on_accept: on_accept()
            if on_close: on_close()

        def cancel_cb(widget):
            self._close_dialog()
            if on_cancel: on_cancel()
            if on_close: on_close()

        def change_cb(widget):
            if on_change: on_change()

        # Derive history file name from command
        import re
        filename = re.sub('[/\n]', '__', ''.join(cmd))
        history_file_base = os.path.join(objects.localcfg['tui.cli.history-dir'].full_path, filename)

        columns_args = [('pack', urwid.Text(':'))]
        self._cmd_parts = []
        self._edit_widgets = []
        edit_index = 0
        for part in cmd:
            if part[0] == '[' and part[-1] == ']':
                edit_index += 1
                history_file = history_file_base + '.input%d' % edit_index
                log.debug('History file for edit #%d: %r', edit_index, history_file)
                edit_widget = CLIEditWidget(on_change=change_cb,
                                            on_accept=accept_cb,
                                            on_cancel=cancel_cb,
                                            history_file=history_file)
                edit_widget.edit_text = part[1:-1]
                edit_widget.edit_pos = len(edit_widget.edit_text)
                columns_args.append(urwid.AttrMap(edit_widget, 'prompt'))
                self._cmd_parts.append(edit_widget)
                self._edit_widgets.append(edit_widget)
            else:
                columns_args.append(('pack', urwid.Text(part)))
                self._cmd_parts.append(part)

        class MyColumns(urwid.Columns):
            """Use <tab> and <shift-tab> to move focus between input fields"""
            def keypress(self, size, key):

                # We do not need to check for division by zero, because this
                # widget is always constructed non-empty (the command exits
                # before calling _open_dialog otherwise)
                def move_right():
                    self.focus_position = (self.focus_position + 1) % len(self.contents)

                def move_left():
                    self.focus_position = (self.focus_position - 1) % len(self.contents)

                if key == 'tab':
                    move_right()
                    while not isinstance(self.focus.base_widget, CLIEditWidget):
                        move_right()
                elif key == 'shift-tab':
                    move_left()
                    while not isinstance(self.focus.base_widget, CLIEditWidget):
                        move_left()
                else:
                    log.debug('focus pos: %r', self.focus_position)
                    return super().keypress(size, key)

        columns_widget = MyColumns(columns_args)

        # Close any previously opened dialog
        from ...tui.tuiobjects import widgets
        if widgets.exists(self._WIDGET_NAME):
            self._close_dialog()

        # Focus the first empty input widget if there are any
        for i,(w,_) in enumerate(columns_widget.contents):
            w = w.base_widget
            log.debug('%02d: %r', i, w)
            if hasattr(w, 'edit_text') and w.edit_text == '':
                columns_widget.focus_position = i
                break

        widgets.add(name=self._WIDGET_NAME,
                    widget=urwid.AttrMap(columns_widget, 'cli'),
                    position=widgets.get_position('cli'),
                    removable=True,
                    options='pack')

    def _close_dialog(self):
        from ...tui.tuiobjects import widgets
        widgets.remove(self._WIDGET_NAME)
        widgets.focus_name = 'main'

    def _run_cmd_or_open_dialog(self, cmd):
        if not cmd:
            return
        elif len(cmd) == 1:
            log.debug('Running command without dialog: %r', cmd)
            self._run_cmd(cmd[0])
        else:
            log.debug('Running command in dialog: %r', cmd)
            self._open_dialog(cmd, on_accept=self._run_cmd_from_dialog)

    def _run_cmd_from_dialog(self):
        cmd = []
        for part in self._cmd_parts:
            if hasattr(part, 'edit_text'):
                cmd.append(part.edit_text)
            else:
                cmd.append(part)
        cmd = ''.join(cmd)
        log.debug('Got command from current dialog: %r', cmd)
        self._run_cmd(cmd)

    def _run_cmd(self, cmd):
        log.debug('Running cmd: %r', cmd)
        if self._ignore_errors:
            # Overload the error() method on the command's instance
            objects.cmdmgr.run_task(cmd, error=lambda msg: None)
        else:
            objects.cmdmgr.run_task(cmd)

    async def _parse_cmd(self, cmd):
        assert isinstance(cmd, str)
        args = await self.parse_placeholders(cmd)
        return self._split_cmd_at_inputs(args[0])

    def _split_cmd_at_inputs(self, cmd):
        """
        Split `cmd` so that each input marker ("[...]") is a single item

        Example result:
            ['somecmd --an-argument ', '[user input goes here]', ' some more arguments']
        """

        log.debug('Splitting %r', cmd)
        parts = [part for part in self._input_regex.split(cmd) if part]
        log.debug('Split: %r', parts)

        for i in range(len(parts)):
            parts[i] = parts[i].replace('\\[', '[')
        log.debug('Unescaped: %r', parts)

        return parts


class MarkCmd(metaclass=CommandMeta):
    name = 'mark'
    provides = {'tui'}
    category = 'tui'
    description = 'Select torrents or files for an action'
    usage = ('mark [<OPTIONS>]',)
    argspecs = (
        {'names': ('--focus-next','-n'), 'action': 'store_true',
         'description': 'Move focus forward after marking or toggling'},
        {'names': ('--toggle','-t'), 'action': 'store_true',
         'description': 'Mark if unmarked, unmark if marked'},
        {'names': ('--all','-a'), 'action': 'store_true',
         'description': 'Mark or toggle all items'},
    )
    more_sections = {
        'NOTES': (('The column "marked" must be in the "columns.*" settings. Otherwise '
                   'marked list items are indistinguishable from unmarked ones.'),
                  '',
                  ('The character that is displayed in the "marked" column is '
                   'specified by the settings "tui.marked.on" and "tui.marked.off".')),
    }

    def run(self, focus_next, toggle, all):
        from ...tui.tuiobjects import tabs
        widget = tabs.focus
        if not widget.has_marked_column:
            raise CmdError('Nothing to mark here.')
        else:
            widget.mark(toggle=toggle, all=all)
            if focus_next:
                widget.focus_position += 1


class UnmarkCmd(metaclass=CommandMeta):
    name = 'unmark'
    provides = {'tui'}
    category = 'tui'
    description = 'Deselect torrents or files for an action'
    usage = ('unmark [<OPTIONS>]',)
    argspecs = (
        {'names': ('--focus-next','-n'), 'action': 'store_true',
         'description': 'Move focus forward after unmarking or toggling'},
        {'names': ('--toggle','-t'), 'action': 'store_true',
         'description': 'Mark if unmarked, unmark if marked'},
        {'names': ('--all','-a'), 'action': 'store_true',
         'description': 'Unmark or toggle all items'},
    )
    more_sections = MarkCmd.more_sections

    def run(self, focus_next, toggle, all):
        from ...tui.tuiobjects import tabs
        widget = tabs.focus
        if not widget.has_marked_column:
            raise CmdError('Nothing to unmark here.')
        else:
            widget.unmark(toggle=toggle, all=all)
            if focus_next:
                widget.focus_position += 1


class QuitCmd(metaclass=CommandMeta):
    name = 'quit'
    provides = {'tui'}
    category = 'tui'
    description = 'Terminate the TUI'

    def run(self):
        import urwid
        raise urwid.ExitMainLoop()


class FindCmd(metaclass=CommandMeta):
    name = 'find'
    provides = {'tui'}
    category = 'tui'
    description = 'Find text in the content of the focused tab'
    usage = ('find [<OPTIONS>] [<PHRASE>]',)
    argspecs = (
        {'names': ('--clear','-c'), 'action': 'store_true',
         'description': ('Remove previously applied filter; this is '
                         'the default if no PHRASE arguments are provided')},
        {'names': ('--next','-n'), 'action': 'store_true',
         'description': 'Jump to next match (call `find <PHRASE>` first)'},
        {'names': ('--previous','-p'), 'action': 'store_true',
         'description': 'Jump to previous match (call `find <PHRASE>` first)'},
        {'names': ('PHRASE',), 'nargs': '*',
         'description': 'Search phrase'},
    )

    def run(self, clear, next, previous, PHRASE):
        from ...tui.tuiobjects import tabs
        content = tabs.focus.base_widget
        if not hasattr(content, 'search_phrase'):
            raise CmdError('This tab does not support finding.')
        elif next and previous:
            raise CmdError('The options --next and --previous contradict each other.')
        elif next:
            if content.search_phrase is None:
                raise CmdError('Set a search phrase first with `find <PHRASE>`.')
            else:
                content.jump_to_next_match()
        elif previous:
            if content.search_phrase is None:
                raise CmdError('Set a search phrase first with `find <PHRASE>`.')
            else:
                content.jump_to_prev_match()
        elif clear:
            content.search_phrase = None
        else:
            try:
                content.search_phrase = ' '.join(PHRASE)
                content.maybe_jump_to_next_match()
            except ValueError as e:
                raise CmdError(e)


class LimitCmd(metaclass=CommandMeta):
    name = 'limit'
    provides = {'tui'}
    category = 'tui'
    description = 'Limit contents of the focused tab by applying more filters'
    usage = ('limit [<OPTIONS>] [<FILTER> <FILTER> ...]',)
    argspecs = (
        {'names': ('--clear','-c'), 'action': 'store_true',
         'description': ('Remove previously applied filter; this is '
                         'the default if no FILTER arguments are provided')},
        {'names': ('FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filters`)'},
    )

    def run(self, clear, FILTER):
        from ...tui.tuiobjects import tabs
        content = tabs.focus.base_widget
        if not hasattr(content, 'secondary_filter'):
            raise CmdError('This tab does not support limiting.')
        else:
            if clear or not FILTER:
                content.secondary_filter = None
            else:
                try:
                    content.secondary_filter = FILTER
                except ValueError as e:
                    raise CmdError(e)

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        from ...tui.tuiobjects import tabs
        from ...tui.views import (FileListWidget, PeerListWidget, SettingListWidget,
                                  TorrentListWidget, TrackerListWidget)
        widget = tabs.focus.base_widget
        if hasattr(widget, 'secondary_filter'):
            if isinstance(widget, TorrentListWidget):
                return candidates.torrent_filter(args.curarg)
            elif isinstance(widget, FileListWidget):
                torrent_filter = 'id=%s' % (widget.focused_torrent_id,)
                return candidates.file_filter(args.curarg, torrent_filter)
            elif isinstance(widget, PeerListWidget):
                return candidates.peer_filter(args.curarg, None)
            elif isinstance(widget, TrackerListWidget):
                torrent_filter = '|'.join('id=%s' % (itemw.torrent_id,)
                                          for itemw in widget.items)
                return candidates.tracker_filter(args.curarg, torrent_filter)
            elif isinstance(widget, SettingListWidget):
                return candidates.setting_filter(args.curarg)


class SortCmd(metaclass=CommandMeta):
    name = 'sort'
    aliases = ()
    provides = {'tui'}
    category = 'tui'
    description = "Sort lists of torrents/peers/trackers/etc"
    usage = ('sort [<OPTIONS>] [<ORDER> <ORDER> <ORDER> ...]',)
    examples = ('sort tracker status !rate-down',
                'sort --add eta')
    argspecs = (
        {'names': ('ORDER',), 'nargs': '*',
         'description': 'How to sort list items (see SORT ORDERS section)'},

        {'names': ('--add', '-a'), 'action': 'store_true',
         'description': 'Append ORDERs to current list of sort orders instead of replacing it'},

        {'names': ('--delete', '-d'), 'action': 'store_true',
         'description': 'Delete ORDERs from current list of sort orders instead of replacing it'},

        {'names': ('--reset', '-r'), 'action': 'store_true',
         'description': 'Go back to sort order that was used when the list was created'},

        {'names': ('--none', '-n'), 'action': 'store_true',
         'description': 'Remove all sort orders from the list'},
    )

    def _list_sort_orders(title, sortcls):
        return (title,) + \
            tuple('\t{}\t - \t{}'.format(', '.join((sname,) + s.aliases), s.description)
                  for sname,s in sorted(sortcls.SORTSPECS.items()))

    more_sections = {
        'SORT ORDERS': (_list_sort_orders('TORRENT LISTS', client.TorrentSorter) +
                        ('',) +
                        _list_sort_orders('PEER LISTS', client.PeerSorter) +
                        ('',) +
                        _list_sort_orders('TRACKER LISTS', client.TrackerSorter))
    }

    async def run(self, add, delete, reset, none, ORDER):
        from ...tui.tuiobjects import tabs
        current_tab = tabs.focus.base_widget

        if reset:
            current_tab.sort = 'RESET'

        if none:
            current_tab.sort = None

        if ORDER:
            # # Find appropriate sorter class for focused list
            sortcls = self._widget2sortcls(current_tab)
            if sortcls is None:
                raise CmdError('Current tab is not sortable.')
            try:
                new_sort = sortcls(utils.listify_args(ORDER))
            except ValueError as e:
                raise CmdError(e)

            if add and current_tab.sort is not None:
                current_tab.sort += new_sort
            elif delete and current_tab.sort is not None:
                current_tab.sort -= new_sort
            else:
                current_tab.sort = new_sort

    @staticmethod
    def _widget2sortcls(list_widget):
        from ...tui.views import (PeerListWidget, SettingListWidget, TorrentListWidget,
                                  TrackerListWidget)
        if isinstance(list_widget, TorrentListWidget):
            return client.TorrentSorter
        elif isinstance(list_widget, PeerListWidget):
            return client.PeerSorter
        elif isinstance(list_widget, TrackerListWidget):
            return client.TrackerSorter
        elif isinstance(list_widget, SettingListWidget):
            return client.SettingSorter

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        from ...tui.tuiobjects import tabs
        sortcls = cls._widget2sortcls(tabs.focus.base_widget)
        if sortcls is not None:
            return candidates.sort_orders(sortcls.__name__)


class TabCmd(mixin.select_torrents, metaclass=CommandMeta):
    name = 'tab'
    provides = {'tui'}
    category = 'tui'
    description = 'Open, close and focus tabs'
    usage = ('tab [<OPTIONS>]',
             'tab [<OPTIONS>] <COMMAND>')
    examples = ('tab',
                'tab -c',
                'tab -c active',
                'tab ls active',
                'tab -b ls active',
                'tab -f active',
                'tab -f 3 ls active',
                'tab -b -f -1 ls active')
    argspecs = (
        {'names': ('--background', '-b'), 'action': 'store_true',
         'description': 'Do not focus new tab'},
        {'names': ('--close-all', '-C'), 'action': 'store_true',
         'description': 'Close all tabs'},
        {'names': ('--close', '-c'), 'nargs': '?', 'default': False, 'document_default': False,
         'description': 'Close focused or specified tab (see TAB IDENTIFIERS SECTION)'},
        {'names': ('--focus', '-f'),
         'description': 'Focus specified tab (see TAB IDENTIFIERS SECTION)'},
        {'names': ('--move', '-m'),
         'description': 'Move focused tab left, right or to absolute position'},
        {'names': ('--title', '-t'),
         'description': 'Manually set tab title instead of generating one'},
        {'names': ('COMMAND',), 'nargs': 'REMAINDER',
         'description': ('Command to run in tab')},
    )
    more_sections = {
        'TAB IDENTIFIERS': (
            'There are three ways to specify a tab (e.g. to close it):',
            ('  - \tIntegers specify the position of the tab.  Positive numbers '
             'start from the left and negative numbers start from the right '
             '(1 (and 0) is the leftmost tab and -1 is the rightmost tab).'),
            ('  - \t"left" and "right" specify the tabs next to the '
             'currently focused tab.'),
            ('  - \tAnything else is assumed to be a part of a tab title.  If there '
             'are multiple matches, the first match from the left wins.'),
        ),
    }

    async def run(self, background, close_all, close, focus, move, title, COMMAND):
        from ...tui.tuiobjects import tabs
        tabid_old = tabs.get_id()

        # Find relevant tab IDs and fail immediately if unsuccessful
        if focus is not None:
            tabid_focus = self._get_tab_id(focus)
            if tabid_focus is None:
                raise CmdError('No such tab: %r' % (focus,))
        if close is not False:
            tabid_close = self._get_tab_id(close)
            if tabid_close is None:
                if close is None:
                    raise CmdError('No tab is open')
                else:
                    raise CmdError('No such tab: %r' % (close,))

        # COMMAND may get additional hidden arguments as instance attributes
        cmd_attrs = {}

        # Apply close/focus/move operations
        if focus is not None:
            log.debug('Focusing tab %r', tabid_focus)
            tabs.focus_id = tabid_focus
        if close_all is not False:
            log.debug('Closing all tabs')
            tabs.clear()
        elif close is not False:
            log.debug('Closing tab %r', tabid_close)
            tabs.remove(tabid_close)
        elif move and tabs.focus:
            self._move_tab(tabs, move)

        # If no tabs were closed, focused or moved, open a new one
        if close is False and close_all is False and focus is None and not move:
            titlew = make_tab_title_widget(title or 'Empty tab',
                                           attr_unfocused='tabs.unfocused',
                                           attr_focused='tabs.focused')
            tabs.insert(titlew, position='right')
            log.debug('Inserted new tab at position %d: %r', tabs.focus_position, titlew.base_widget.text)

        # Maybe provide a user-specified tab title to the new command
        if title:
            cmd_attrs['title'] = title

        if COMMAND:
            # Execute command
            cmd_str = ' '.join(shlex.quote(arg) for arg in COMMAND)
            log.debug('Running command in tab %s with args %s: %r',
                      tabs.focus_position,
                      ', '.join('%s=%r' % (k,v) for k,v in cmd_attrs.items()),
                      cmd_str)

            success = await objects.cmdmgr.run_async(cmd_str, **cmd_attrs)
        else:
            success = True

        if background:
            tabs.focus_id = tabid_old
        else:
            content = tabs.focus
            if content is not None and hasattr(content, 'marked_count'):
                from ...tui.tuiobjects import bottombar
                bottombar.marked.update(content.marked_count)

        return success

    def _get_tab_id(self, pos):
        from ...tui.tuiobjects import tabs
        if len(tabs) == 0:
            return None

        if pos is None:
            return tabs.focus_id

        def find_id_by_index(index):
            try:
                index = int(index)
            except ValueError:
                pass
            else:
                index_max = len(tabs) - 1
                # Internally, first tab is at index 0, but for users it's 1, unless
                # they gave us 0, in which case we assume they mean 1.
                index = index - 1 if index > 0 else index

                # Limit index to index_max, considering negative values when
                # indexing from the right.
                if index < 0:
                    index = max(index, -index_max - 1)
                else:
                    index = min(index, index_max)
                return tabs.get_id(index)

        def find_right_left_id(right_or_left):
            tabcount = len(tabs)
            if tabcount > 1:
                cur_index = tabs.focus_position
                cur_index = 1 if cur_index is None else cur_index
                if right_or_left == 'left':
                    return tabs.get_id(max(0, cur_index - 1))
                elif right_or_left == 'right':
                    return tabs.get_id(min(tabcount - 1, cur_index + 1))

        def find_id_by_title(string):
            for index,title in enumerate(tabs.titles):
                if string in title.original_widget.text:
                    return tabs.get_id(index)

        # Try to use pos as an index
        tabid = find_id_by_index(pos)
        if tabid is not None:
            log.debug('Found tab ID by index: %r -> %r', pos, tabid)
            return tabid

        pos_str = str(pos)

        # Move to left/right tab
        tabid = find_right_left_id(pos_str)
        if tabid is not None:
            log.debug('Found tab ID by direction: %r -> %r', pos, tabid)
            return tabid

        # Try to find tab title
        tabid = find_id_by_title(pos_str)
        if tabid is not None:
            log.debug('Found tab ID by title: %r -> %r', pos, tabid)
            return tabid

    def _move_tab(self, tabs, move):
        if move == 'left':
            tabs.move(tabs.get_id(), 'left')
        elif move == 'right':
            tabs.move(tabs.get_id(), 'right')
        else:
            try:
                index = int(move)
            except (ValueError, TypeError):
                raise CmdError('--move argument must be "left", "right" or tab index: %r' % (move,))
            else:
                # Positive tab index starts at 0, negative at -1
                if index > 0:
                    index -= 1
                tabs.move(tabs.get_id(), index)

    _own_options = {('--close', '-c'): 1,
                    ('--focus', '-f'): 1,
                    ('--title', '-t'): 1}

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        posargs = args.posargs(cls._own_options)
        if posargs.curarg_index == 1:
            # First positional argument is the subcmd's name
            return candidates.commands()
        else:
            # Any other positional arguments will be passed to subcmd
            subcmd = cls._get_subcmd(args)
            if subcmd:
                return candidates.for_args(subcmd)

    @classmethod
    def completion_candidates_opts(cls, args):
        """Return candidates for arguments that start with '-'"""
        subcmd = cls._get_subcmd(args)
        if subcmd:
            # Get completion candidates for subcmd
            return candidates.for_args(subcmd)
        else:
            # Parent class generates candidates for our own options
            return super().completion_candidates_opts(args)

    @classmethod
    def completion_candidates_params(cls, option, args):
        """Complete parameters (e.g. --option parameter1,parameter2)"""
        if option in ('--close', '--focus'):
            return candidates.tab_titles()

    @classmethod
    def _get_subcmd(cls, args):
        # First posarg is 'tab'
        subcmd_start = args.nth_posarg_index(2, cls._own_options)
        # Subcmd is only relevant if the cursor is somewhere on it.
        # Otherwise, we're on our own arguments.
        if subcmd_start is not None and subcmd_start < args.curarg_index:
            return args[subcmd_start:]


class TUICmd(metaclass=CommandMeta):
    name = 'tui'
    provides = {'tui'}
    category = 'tui'
    description = 'Show or hide parts of the text user interface'
    usage = ('tui <ACTION> <ELEMENT> <ELEMENT> ...',)
    examples = ('tui toggle log',
                'tui hide topbar.help')
    argspecs = (
        {'names': ('ACTION',), 'choices': ('show', 'hide', 'toggle'),
         'description': '"show", "hide" or "toggle"'},
        {'names': ('ELEMENT',), 'nargs': '+',
         'description': ('Name of TUI elements; '
                         'see ELEMENT NAMES section for a list')},
    )
    # HelpManager supports sequences of lines or a callable that returns them
    more_sections = {'ELEMENT NAMES': lambda: ('Available TUI element names are: ' +
                                               ', '.join(_tui_element_names()),)}

    def run(self, ACTION, ELEMENT):
        from ...tui.tuiobjects import widgets
        widget = None
        success = True
        for element in utils.listify_args(ELEMENT):
            # Resolve path
            path = element.split('.')
            target_name = path.pop(-1)
            current_path = []
            widget = widgets
            try:
                for widgetname in path:
                    current_path.append(widgetname)
                    widget = getattr(widget, widgetname)
            except AttributeError:
                self.error('Unknown TUI element: %r' % ('.'.join(current_path),))

            if widget is not None:
                action = getattr(widget, ACTION)
                if any(ACTION == x for x in ('hide', 'toggle')):
                    action = partial(action, free_space=False)

                log.debug('%sing %s in %s', ACTION.capitalize(), target_name, widget)
                try:
                    action(target_name)
                except ValueError as e:
                    success = False
                    self.error(e)
                else:
                    success = success and True

        if not success:
            raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        posargs = args.posargs()
        if posargs.curarg_index == 1:
            for argspec in cls.argspecs:
                if 'ACTION' in argspec['names']:
                    return candidates.Candidates(argspec['choices'],
                                                 label='Action')
        else:
            return candidates.Candidates(_tui_element_names(),
                                         label='Element')

# Lazily load element names from tui module to avoid importing TUI stuff if possible
@functools.lru_cache()
def _tui_element_names():
    from ...tui import tuiobjects
    return tuple(str(name) for name in sorted(tuiobjects.widgets.names_recursive))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""
Application-wide instances that are always needed, regardless of interface
or features
"""

from . import logging, settings
from .client import API
from .commands import CommandManager
from .helpmgr import HelpManager

log = logging.make_logger()

# Set by main.run()
main_rcfile = None

localcfg = settings.Settings()
settings.init_defaults(localcfg)

srvapi = API(interval=localcfg['tui.poll'])

remotecfg = settings.RemoteSettings(srvapi.settings)

cfg = settings.CombinedSettings(localcfg, remotecfg)

helpmgr = HelpManager()

cmdmgr = CommandManager(info_handler=lambda msg: log.info(msg),
                        error_handler=lambda msg: log.error(msg))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from .usertypes import Float, Int

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


class DataSizeConverter():
    """
    Convert bits to bytes or vice versa, ensuring an Int instance with an
    application-wide unit and unit prefix
    """
    _short = {'bit': 'b', 'byte': 'B'}

    def __init__(self):
        self.prefix = 'metric'
        self.unit = 'B'

    def __call__(self, num, unit=None):
        """
        Make Int from `num`

        The returned Int is converted to bits or bytes depending on what the
        `unit` property is set to.

        If no unit is given (by passing a Int/Float object with a specified
        `unit` property or by passing the `unit` argument), it is assumed to be
        what the `unit` property of this object is set to.
        """
        if not isinstance(num, (Int, Float)):
            # Parse unit; fall back to given or our own unit
            num = Int(num, unit=unit or self._unit, prefix=self._prefix)

        unit_given = self._short.get(num.unit, num.unit) or self._unit
        if unit_given not in ('b', 'B'):
            raise ValueError("Unit must be 'b' (bit) or 'B' (byte), not %r" % unit_given)
        else:
            return Int(num, unit=unit_given, convert_to=self._unit, prefix=self._prefix)

    @property
    def unit(self):
        """'b' (bits) or 'B' (bytes)"""
        return self._unit

    @unit.setter
    def unit(self, unit):
        if unit not in ('bit', 'byte', 'b', 'B'):
            raise ValueError("Unit must be 'bit or 'byte'")
        else:
            self._unit = self._short.get(unit, unit)

    @property
    def prefix(self):
        """'binary' or 'metric'"""
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        if prefix not in ('binary', 'metric'):
            raise ValueError("Prefix must be 'binary' or 'metric'")
        else:
            self._prefix = prefix



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import datetime
import operator
import os
import re
import subprocess
import sys
import textwrap
from collections import abc

from ... import __appname__, __version__, objects
from ...client import ClientError
from ...completion import candidates
from ...settings import defaults, rcfile
from ...utils import cached_property, cliparser, string, usertypes
from .. import CmdError, CommandMeta, utils
from . import _mixin as mixin
from ._common import (make_COLUMNS_doc, make_SCRIPTING_doc, make_SORT_ORDERS_doc,
                      make_X_FILTER_spec)

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


class DumpCmdbase(mixin.get_rc_filepath,
                  metaclass=CommandMeta):
    name = 'dump'
    aliases = ()
    category = 'configuration'
    provides = set()
    description = 'Generate commands that reproduce current settings, keybindings and tabs'
    usage = ('dump [<OPTIONS>] [<FILE>]',)
    examples = ('dump rc.current    # Write %s' % (string.tildify(os.path.join(
                os.path.dirname(defaults.DEFAULT_RCFILE), 'rc.current')),),
                'dump .%src.current  # Write rc.current in current working directory' % (os.sep,))
    argspecs = (
        {'names': ('FILE',), 'nargs': '?', 'default': objects.main_rcfile,
         'description': ('Path to rc file or "-" to print to stdout; if FILE does not exist and '
                         'does not start with "{sep}", ".{sep}" or "~", "{dir}" is prepended'.format(
                             sep=os.sep, dir=string.tildify(os.path.dirname(defaults.DEFAULT_RCFILE))))},
        {'names': ('--force', '-f'), 'action': 'store_true',
         'description': 'Overwrite FILE if it exists'},
    )

    DUMP_WIDTH = 79

    @staticmethod
    def _make_header(appname=__appname__, version=__version__, timestamp=None):
        if timestamp is None:
            now = datetime.datetime.now()
            if sys.hexversion < 0x03060000:
                timestamp = now.replace(microsecond=0).isoformat(sep=' ')
            else:
                timestamp = now.isoformat(sep=' ', timespec='minutes')
        return ('# This is an autogenerated rc file for %s version %s.' % (appname, version),
                '# This file was created on %s.' % (timestamp,),
                '# NOTE: If this header exists %s will overwrite this file without warning.' % (appname,))

    @classmethod
    def _make_header_regex(cls):
        lines = cls._make_header(appname=r'\S+', version=r'\S+', timestamp=r'\S+ \S+')
        lines = (line.replace('.', r'\.') for line in lines)
        x = re.compile('^' + '\n'.join(lines) + '\n', flags=re.DOTALL)
        log.debug(x)
        return x

    def run(self, force, FILE):
        lines = self._make_header() + (
            '', '',
            '### TABS',
            '',
            self._get_tabs(),
            '', '',
            '### SETTINGS',
            '',
            self._get_settings(),
            '', '',
            '### KEYBINDINGS',
            '',
            self._get_keybindings(),
        )
        content = '\n'.join(lines)

        if FILE == '-':
            return self.dump_rc(content, force=force, path=None)
        # We can't rely on the default value in tests because objects.main_rcfile is not
        # set in tests.
        elif FILE is None:
            return self.dump_rc(content, force=force, path=objects.main_rcfile)
        else:
            return self.dump_rc(content, force=force, path=self.get_rc_filepath(FILE))

    def write_rc_file(self, content, path, force=False):
        if self._check_write(path, force):
            try:
                with open(path, 'w') as f:
                    f.write(content + '\n')
            except OSError as e:
                raise CmdError('Unable to write %s: %s' % (string.tildify(path), e.strerror))
            else:
                self.info('Wrote rc file: %s' % (string.tildify(path),))
                return True

    def _check_write(self, path, force):
        if not os.path.exists(path):
            return True
        elif force:
            return True
        elif os.path.normpath(path) == os.path.normpath(objects.main_rcfile):
            # If this is the main rcfile and it was written by us, we can overwrite it
            try:
                with open(path, 'r') as f:
                    header = f.read(1024)
            except OSError as e:
                raise CmdError('Unable to read %s: %s' % (string.tildify(path), e.strerror))
            else:
                header_regex = self._make_header_regex()
                if header_regex.search(header):
                    return True
                else:
                    raise CmdError('Not overwriting edited rc file: %s' % (string.tildify(path),))
        raise CmdError('File exists: %s' % (string.tildify(path),))

    def _get_settings(self):
        lcfg = objects.localcfg
        settings = []
        for name in sorted(lcfg):
            value = lcfg[name]
            if isinstance(value, abc.Iterable) and not isinstance(value, str):
                value = ' '.join(value)
            else:
                value = str(value)
                if ' ' in value:
                    value = repr(value)
            escape_set_cmd = lcfg[name] == lcfg.default(name)
            lines = (self._wrap_setting_description(lcfg.description(name)) +
                     self._wrap_setting_default(lcfg.default(name)) +
                     self._wrap_set_cmd(name, value, escape=escape_set_cmd))
            settings.append('\n'.join(lines))
        return '\n\n'.join(settings)

    def _wrap_setting_description(self, string):
        return textwrap.wrap('# ' + str(string),
                             width=self.DUMP_WIDTH,
                             subsequent_indent='# ')

    def _wrap_setting_default(self, string):
        subseqind = '# ' + len('Default: ') * ' '
        prefix = '# Default: '
        lines = textwrap.wrap(prefix + str(string),
                              width=self.DUMP_WIDTH,
                              subsequent_indent=subseqind,
                              break_long_words=False,
                              break_on_hyphens=False)
        # First line must always contain the beginning of the default value, even if it is
        # too long
        if len(lines) >= 2 and lines[0].strip() == prefix.strip():
            lines[0] = lines[0] + lines[1][len(subseqind) - 1:]
            del lines[1]
        return lines

    def _wrap_set_cmd(self, name, value, escape):
        cmd = 'set'
        lines = textwrap.wrap(' '.join((cmd, name, value)),
                              width=self.DUMP_WIDTH,
                              subsequent_indent=(len(cmd) + len(name) + 2) * ' ',
                              break_long_words=False,
                              break_on_hyphens=False)

        # First line must always contain command, name and beginning of value, even if it
        # is too long
        if len(lines) >= 3 and lines[0].strip() == cmd:
            lines[0] = ' '.join((lines[0], lines[1].strip(), lines[2].strip()))
            del lines[1], lines[1]
        elif len(lines) >= 2 and lines[0].strip() == ' '.join((cmd, name)):
            lines[0] = ' '.join((lines[0], lines[1].strip()))
            del lines[1]

        # Escape line breaks of multi-line commands
        for i in range(len(lines) - 1):
            lines[i] += ' \\'

        if escape:
            return ['#' + line for line in lines]
        else:
            return lines

    def _get_keybindings(self):
        # Calls to "bind" are ignored in CLI mode
        if objects.cmdmgr.active_interface != 'tui':
            with objects.cmdmgr.temporary_active_interface('tui'):
                objects.cmdmgr.run_ignored_calls_sync('bind')

        from ...tui.tuiobjects import keymap
        contexts = []
        for context in sorted(keymap.contexts):
            # Command can consist of one or two lines
            lines = []
            for key,action in keymap.map(context):
                desc = keymap.get_description(key, context)
                escape = self._is_default_keybinding(key, context, action, desc)
                cmd = self._wrap_bind_cmd(key, action, context, desc, escape)
                lines.extend(cmd)
            contexts.append('\n'.join(lines))
        return '\n\n'.join(contexts)

    def _is_default_keybinding(self, key, context, action, description):
        from ...tui.tuiobjects import keymap
        dct = {'key': key, 'action': action}
        if context != keymap.DEFAULT_CONTEXT:
            dct['context'] = context
        if description:
            dct['description'] = description
        for d in self._default_keymap:
            if dct == d:
                return True
        return False

    @cached_property
    def _default_keymap(self):
        # Convert string keys to Key objects
        from ...tui.tuiobjects import keymap
        km = []
        for kwargs in defaults.DEFAULT_KEYMAP:
            km.append({**kwargs, **{'key': keymap.mkkey(kwargs['key'])}})
        return km

    def _wrap_bind_cmd(self, key, action, context, description, escape=False):
        from ...tui.tuiobjects import keymap
        lines = [['bind']]
        if description:
            lines[0].extend(('--description', cliparser.quote(description), '\\'))
            indent = len(lines[0][0]) * ' '
            lines.append([indent])
        if context != keymap.DEFAULT_CONTEXT:
            lines[-1].extend(('--context', cliparser.quote(context))),
        lines[-1].append(cliparser.quote(str(key)))
        lines[-1].append(str(action))
        if escape:
            return ['#' + ' '.join(line) for line in lines]
        else:
            return [' '.join(line) for line in lines]

    def _get_tabs(self):
        return '\n'.join(self._get_tab_cmds())

    def _get_tab_cmds(self):
        # Calls to "tab" are ignored in CLI mode
        if objects.cmdmgr.active_interface != 'tui':
            with objects.cmdmgr.temporary_active_interface('tui'):
                objects.cmdmgr.run_ignored_calls_sync('tab')

        # Get commands from opened tabs
        from ...tui.tuiobjects import tabs
        cmds = []
        for tabid in tabs.ids:
            cmd = tabs.get_info(tabid).get('command', None)
            if cmd is not None:
                cmds.append('tab ' + cmd)

        if not cmds:
            # No tabs were opened by the user
            cmds = list(defaults.DEFAULT_TAB_COMMANDS)

        # Store current tab
        if tabs.focus_position is not None:
            cmds.append('tab --focus %d' % (tabs.focus_position + 1,))

        return cmds


class RcCmdbase(mixin.get_rc_filepath,
                metaclass=CommandMeta):
    name = 'rc'
    aliases = ('source',)
    category = 'configuration'
    provides = set()
    description = 'Run commands in rc file'
    usage = ('rc <FILE>',)
    examples = ('rc rc.example    # Load $XDG_CONFIG_HOME{sep}.config{sep}{{__appname__}}{sep}rc.example'.format(sep=os.sep),
                'rc .{sep}rc.example  # Load rc.example from current working directory'.format(sep=os.sep))
    argspecs = (
        {'names': ('FILE',),
         'description': ('Path to rc file; if FILE does not exist and does not start with '
                         '"{sep}", ".{sep}" or "~", '
                         '"$XDG_CONFIG_HOME{sep}.config{sep}{{__appname__}}{sep}" '
                         'is prepended').format(sep=os.sep)},
    )

    async def run(self, FILE):
        filepath = self.get_rc_filepath(FILE)
        try:
            lines = rcfile.read(filepath)
        except rcfile.RcFileError as e:
            raise CmdError('Loading rc file failed: %s' % e)
        else:
            log.debug('Running commands from rc file: %r', filepath)
            for cmdline in lines:
                success = await objects.cmdmgr.run_async(cmdline)
                # False means failure, None means the command didn't run because
                # the active interface doesn't support it
                if success is False:
                    raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        # Command only takes one argument
        if args.curarg_index == 1:
            return candidates.fs_path(args.curarg.before_cursor,
                                      base=os.path.dirname(defaults.DEFAULT_RCFILE))


class ResetCmdbase(metaclass=CommandMeta):
    name = 'reset'
    category = 'configuration'
    provides = set()
    description = 'Reset settings to their default values'
    usage = ('reset <NAME> <NAME> <NAME> ...',)
    examples = ('reset connect.port',)
    argspecs = (
        {'names': ('NAME',), 'nargs': '+',
         'description': 'Name of setting'},
    )
    more_sections = {
        'SEE ALSO': ('Run `help settings` for a list of all available settings.',
                     'Note that remote settings (srv.*) cannot be reset.'),
    }

    def run(self, NAME):
        success = True
        for name in utils.listify_args(NAME):
            try:
                objects.cfg.reset(name)
            except NotImplementedError:
                self.error('Remote settings cannot be reset: %s' % name)
                success = False
            except KeyError:
                self.error('Unknown setting: %s' % name)
                success = False
        if not success:
            raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        return candidates.setting_names()


class SetCmdbase(mixin.get_setting_sorter, mixin.get_setting_columns,
                 metaclass=CommandMeta):
    name = 'set'
    category = 'configuration'
    provides = set()
    description = 'Change or list settings'
    usage = ('set [<NAME>[:eval]] [<VALUE>]',)
    examples = ('set connect.host my.server.example.org',
                'set connect.user jonny_sixpack',
                'set connect.password:eval getpw --id transmission',
                'set tui.log.height +=10')
    from ...client.sorters import SettingSorter
    from ...views.setting import COLUMNS
    argspecs = (
        {'names': ('NAME',), 'nargs': '?',
         'description': "Name of setting; append ':eval' to turn VALUE into a shell command"},

        {'names': ('VALUE',), 'nargs': 'REMAINDER',
         'description': ('New value or shell command that prints the new value to stdout; '
                         "numerical values can be adjusted by prepending '+=' or '-='")},

        {'names': ('--sort', '-s'),
         'description': ('Comma-separated list of sort orders when listing settings '
                         '(see SORT ORDERS section)')},

        {'names': ('--columns', '-c'),
         'default_description': "current value of 'columns.settings' setting",
         'description': ('Comma-separated list of column names when listing settings '
                         '(see COLUMNS section)')},
    )
    more_sections = {
        'COLUMNS': make_COLUMNS_doc(COLUMNS, '--columns', 'columns.settings'),
        'SORT ORDERS': make_SORT_ORDERS_doc(SettingSorter, '--sort', 'sort.settings'),
        'SCRIPTING': make_SCRIPTING_doc(name),
        'SEE ALSO': (('Run `help settings` or run `set` without commands '
                      'for a list of available local and remote settings.'),),
    }

    async def run(self, NAME, VALUE, sort, columns):
        if not NAME and not VALUE:
            # Get remote setting values
            try:
                await objects.cfg.update()
            except ClientError as e:
                error = e
            else:
                error = None

            # Show list of settings
            sort = objects.localcfg['sort.settings'] if sort is None else self._parse_value(sort, listify=True)
            columns = objects.localcfg['columns.settings'] if columns is None else self._parse_value(columns, listify=True)
            try:
                sort = self.get_setting_sorter(sort)
                columns = self.get_setting_columns(columns)
            except ValueError as e:
                raise CmdError(e)
            else:
                self.make_setting_list(sort, columns)
                if error:
                    raise CmdError(error)
            return

        # NAME might have ':eval' attached if VALUE is shell command.
        try:
            name = self._parse_name(NAME)
        except ValueError as e:
            raise CmdError(e)

        # Get current value in case we want to display or adjust it
        if objects.cfg.is_remote(name):
            try:
                await objects.cfg.update()
            except ClientError as e:
                raise CmdError(e)

        # VALUE might be shell command or have '+='/'-=' prepended
        try:
            value = self._parse_value(VALUE,
                                      listify=isinstance(objects.cfg[name], (tuple, list)),
                                      is_cmd=NAME.endswith(':eval'))
        except ValueError as e:
            # Report potential stderr output if VALUE is a command
            raise CmdError('%s: %s' % (name, e))

        # Separate '+=' or '-=' from value
        try:
            op, value = self._get_operator(value)
        except ValueError as e:
            # Report invalid value after operator (e.g. nan)
            raise CmdError('%s = %s: %s' % (name, self._stringify(value), e))

        # Apply operator to current value
        if op is not None:
            try:
                value = self._adjust_value(objects.cfg[name], op, value)
            except ValueError as e:
                # Report out-of-bounds value
                opfunc = getattr(operator, op)
                unbound = objects.cfg[name].copy(min=-float('inf'), max=float('inf'))
                invalid = opfunc(unbound, value)
                raise CmdError('%s = %s: %s' % (name, self._stringify(invalid), e))

        # Update setting's value
        try:
            await objects.cfg.set(name, value)
        except ValueError as e:
            raise CmdError('%s = %s: %s' % (name, self._stringify(value), e))
        except ClientError as e:
            raise CmdError(e)

    def _parse_name(self, name):
        if name.endswith(':eval'):
            name = name[:-5]
        if name in objects.cfg:
            return name
        else:
            raise ValueError('Unknown setting: %s' % name)

    def _parse_value(self, value, listify=False, is_cmd=False):
        if is_cmd:
            value = [self._eval_cmd(value)]
        if listify:
            return utils.listify_args(value)
        else:
            return ' '.join(value)

    @staticmethod
    def _eval_cmd(cmd):
        if not isinstance(cmd, str):
            cmd = ' '.join(cmd)
        log.debug('Running shell command: %r', cmd)
        proc = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout = proc.stdout.decode('utf-8').strip('\n')
        stderr = proc.stderr.decode('utf-8').strip('\n')
        if stderr:
            raise ValueError(stderr)
        else:
            return stdout

    @staticmethod
    def _get_operator(value):
        if isinstance(value, str):
            stripped = value.strip()
            if len(stripped) >= 3:
                def to_num(string):
                    try:
                        return usertypes.Float(string)
                    except ValueError as e:
                        raise ValueError('%s: %r' % (e, string))
                if stripped[:2] == '+=':
                    return '__add__', to_num(stripped[2:])
                elif stripped[:2] == '-=':
                    return '__sub__', to_num(stripped[2:])
        return None, value

    @staticmethod
    def _adjust_value(current, op, value):
        if isinstance(current, (float, int)):
            if current >= float('inf'):
                current = usertypes.Float(0)
            func = getattr(operator, op)
            value = func(current, value)
        return value

    @staticmethod
    def _stringify(value):
        if not isinstance(value, str) and isinstance(value, abc.Iterable):
            return ', '.join(str(item) for item in value)
        else:
            return str(value)

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        # If --columns or --sort is anywhere, we only display options
        for arg in args:
            if cls.short_options.get(arg, arg) in ('--columns', '--sort'):
                return

        settings = candidates.setting_names()
        if args.curarg_index == 1:
            log.debug('Returning setting names: %r', settings)
            return settings
        elif args.curarg_index >= 2:
            log.debug('Completing values for %r', args[1])
            # Remove command name from command line
            return candidates.setting_values(args[1:])

    @classmethod
    def completion_candidates_opts(cls, args):
        """Return candidates for arguments that start with '-'"""
        # Only complete options or parameters for options if there are no
        # positional arguments (i.e. when the command doesn't look like it's
        # changing a setting).  But positional arguments may also be parameters
        # for --columns or --sort.
        posargs = args.posargs({('--columns', '-c'): 1,
                                ('--sort', '-s'): 1})
        if len(posargs) == 1:
            if args.curarg_index == 1:
                return (super().completion_candidates_opts(args),
                        candidates.setting_names())
            else:
                return super().completion_candidates_opts(args)

    @classmethod
    def completion_candidates_params(cls, option, args):
        """Complete parameters (e.g. --option parameter1,parameter2)"""
        if option == '--sort':
            return candidates.sort_orders('SettingSorter')
        elif option == '--columns':
            return candidates.column_names('settings')


class RateLimitCmdbase(metaclass=CommandMeta):
    name = 'ratelimit'
    aliases = ('rate', 'rl')
    provides = set()
    category = 'configuration'
    description = 'Limit transfer rates per torrent or globally'
    usage = ('ratelimit',
             'ratelimit <DIRECTION>',
             'ratelimit <DIRECTION> <LIMIT>',
             'ratelimit <DIRECTION> <LIMIT> <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('ratelimit up 5Mb',
                'ratelimit down,up 1M global',
                'ratelimit up,dn off "This torrent" size<100MB')
    argspecs = (
        {'names': ('DIRECTION',), 'nargs': '?', 'default': 'up,down',
         'description': 'Any combination of "up", "down" or "dn" separated by a comma'},

        {'names': ('LIMIT',), 'nargs': '?',
         'description': ('Maximum allowed transfer rate (see `help srv.limit.rate.up` '
                         'for the syntax) or "show" to display the current limit')},

        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*',
                           more_text=('"global" to set global limit (same as setting '
                                      "srv.limit.rate.<DIRECTION>); may be omitted in CLI mode "
                                      'for the same effect as specifying "global"')),

        {'names': ('--quiet','-q'), 'action': 'store_true',
         'description': 'Do not show new bandwidth rate(s)'},
    )

    async def run(self, DIRECTION, LIMIT, TORRENT_FILTER, quiet):
        directions = tuple('down' if d == 'dn' else d
                           for d in map(str.lower, DIRECTION.split(',')))
        for d in directions:
            if d not in ('up', 'down'):
                raise CmdError('Invalid direction: %r' % (d,))

        # _show_limits() and _set_limits() are defined in cli.config and
        # tui.config because the TUI can use the focused torrent while the CLI
        # can't.
        if not LIMIT or LIMIT == 'show':
            await self._show_limits(TORRENT_FILTER, directions)
        else:
            # Do we adjust current limits or set absolute limits?
            limit = LIMIT.strip()
            if limit[:2] == '+=' or limit[:2] == '-=':
                adjust = True
                limit = limit[0] + limit[2:]  # Remove '=' so it can be parsed as a number
            else:
                adjust = False

            await self._set_limits(TORRENT_FILTER, directions, limit,
                                   adjust=adjust, quiet=quiet)

    async def _show_global_limits(self, directions):
        for d in directions:
            get_method = getattr(objects.srvapi.settings, 'get_limit_rate_' + d)
            limit = await get_method()
            self._output('Global %sload rate limit: %s' % (d, limit))

    async def _show_individual_limits(self, TORRENT_FILTER, directions):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        request = objects.srvapi.torrent.torrents(
            tfilter, keys=('name', 'limit-rate-up', 'limit-rate-down'))
        response = await self.make_request(request, polling_frenzy=True, quiet=True)
        if response.success:
            for t in response.torrents:
                for d in directions:
                    self._output('%s %sload rate limit: %s' %
                                 (t['name'], d, t['limit-rate-%s' % d]))

    async def _set_global_limits(self, directions, limit, quiet=False, adjust=False):
        for d in directions:
            log.debug('Setting global %s rate limit: %r', d, limit)
            set_method = getattr(objects.srvapi.settings,
                                 ('adjust' if adjust else 'set') + '_limit_rate_' + d)
            get_method = getattr(objects.srvapi.settings, 'get_limit_rate_' + d)
            try:
                try:
                    await set_method(limit)
                except ValueError as e:
                    raise CmdError('%s: %r' % (e, limit))
                if not quiet:
                    limit = await get_method()
                    self.info('Global %sload rate limit: %s' % (d, limit))
            except ClientError as e:
                raise CmdError(e)

    async def _set_individual_limits(self, TORRENT_FILTER, directions, limit, quiet=False, adjust=False):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)

        log.debug('Setting %sload rate limit for %s torrents: %r',
                  '+'.join(directions), tfilter, limit)

        success = True
        for d in directions:
            method = getattr(objects.srvapi.torrent,
                             ('adjust' if adjust else 'set') + '_limit_rate_' + d)
            response = await self.make_request(method(tfilter, limit),
                                               polling_frenzy=True, quiet=quiet)
            success = success and response.success
        if not success:
            raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        posargs = args.posargs()
        if posargs.curarg_index == 1:
            cands = []
            if 'up' not in posargs.curarg:
                cands.append('up')
            if all(x not in posargs.curarg for x in ('down', 'dn')):
                cands.append('down')
            return candidates.Candidates(cands, label='Direction', curarg_seps=(',',))
        elif posargs.curarg_index >= 3:
            return candidates.torrent_filter(args.curarg)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import asyncio
import os
import sys

from . import cliopts, logging, objects, settings
from .objects import cmdmgr, log, srvapi

# Remove python from process name when running inside tmux
if 'TMUX' in os.environ:
    try:
        from setproctitle import setproctitle
    except ImportError:
        pass
    else:
        from . import __appname__
        setproctitle(__appname__)

cliargs, clicmds = cliopts.parse()
objects.main_rcfile = cliargs['rcfile'] or settings.defaults.DEFAULT_RCFILE

logging.setup(debugmods=cliargs['debug'], filepath=cliargs['debug_file'])
logging.redirect_level('INFO', sys.stdout)


def run():
    cmdmgr.load_cmds_from_module('stig.commands.cli', 'stig.commands.tui')

    from . import hooks  # noqa: F401
    from .commands import CmdError
    from .commands.guess_ui import UIGuessError, guess_ui

    # Read commands from rc file
    rclines = ()
    if not cliargs['norcfile']:
        from .settings import rcfile
        try:
            rclines = rcfile.read(objects.main_rcfile)
        except rcfile.RcFileError as e:
            log.error('Loading rc file failed: {}'.format(e))
            sys.exit(1)

    # Decide if we run as a TUI or CLI
    if cliargs['tui']:
        cmdmgr.active_interface = 'tui'
    elif cliargs['notui']:
        cmdmgr.active_interface = 'cli'
    else:
        try:
            cmdmgr.active_interface = guess_ui(clicmds, cmdmgr)
        except UIGuessError:
            log.error('Unable to guess user interface')
            log.error('Provide one of these options: --tui/-t or --no-tui/-T')
            sys.exit(1)
        except CmdError as e:
            log.error(e)
            sys.exit(1)

    def run_commands():
        for cmdline in rclines:
            success = cmdmgr.run_sync(cmdline)
            # Ignored commands return None, which we consider a success here
            # because TUI commands like 'tab' in the rc file should have no
            # effect at all when in CLI mode.
            if success is False:
                return False

        # Exit if CLI commands fail
        if clicmds:
            success = cmdmgr.run_sync(clicmds)
            if not success:
                return False

        return True

    exit_code = 0

    # Run commands either in CLI or TUI mode
    if cmdmgr.active_interface == 'cli':
        # Exit when pipe is closed (e.g. `stig help | head -1`)
        import signal
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)

        try:
            if not run_commands():
                exit_code = 1
        except KeyboardInterrupt:
            log.debug('Caught SIGINT')

    elif cmdmgr.active_interface == 'tui':
        from .tui import main as tui
        if not tui.run(run_commands):
            exit_code = 1

    asyncio.get_event_loop().run_until_complete(srvapi.rpc.disconnect('Quit'))

    # We're not closing the AsyncIO event loop here because it sometimes
    # complains about unfinished tasks and not calling it seems to work fine.
    sys.exit(exit_code)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""TUI and CLI specs for tracker list columns"""

from . import ColumnBase

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


COLUMNS = {}
ALIASES = {'dom'   : 'domain',
           'an'    : 'url-announce',
           'sc'    : 'url-scrape',
           'st'    : 'status',
           'err'   : 'err',
           'erran' : 'error-announce',
           'errsc' : 'error-scrape',
           'dns'   : 'downloads',
           'lcs'   : 'leeches',
           'sds'   : 'seeds',
           'lan'   : 'last-announce',
           'nan'   : 'next-announce',
           'lsc'   : 'last-scrape',
           'nsc'   : 'next-scrape'}


class Torrent(ColumnBase):
    header = {'left': 'Torrent'}
    align = 'left'
    width = None
    min_width = 7
    may_have_wide_chars = True

    def get_value(self):
        return self.data['tname']

COLUMNS['torrent'] = Torrent


class Tier(ColumnBase):
    header = {'left': 'Tier'}
    align = 'right'
    width = 4
    min_width = 4

    def get_value(self):
        return self.data['tier']

COLUMNS['tier'] = Tier


class Domain(ColumnBase):
    header = {'left': 'Domain'}
    align = 'left'
    width = None
    min_width = 5

    def get_value(self):
        return self.data['domain']

COLUMNS['domain'] = Domain


class AnnounceURL(ColumnBase):
    header = {'left': 'Announce'}
    align = 'left'
    width = None
    min_width = 10

    def get_value(self):
        return self.data['url-announce']

COLUMNS['url-announce'] = AnnounceURL


class ScrapeURL(ColumnBase):
    header = {'left': 'Scrape'}
    align = 'left'
    width = None
    min_width = 10

    def get_value(self):
        return self.data['url-scrape']

COLUMNS['url-scrape'] = ScrapeURL


class Status(ColumnBase):
    header = {'left': 'Status'}
    align = 'right'
    width = 10
    min_width = 5

    def get_value(self):
        return self.data['status']

COLUMNS['status'] = Status


class Error(ColumnBase):
    header = {'left': 'Error'}
    align = 'left'
    width = None
    min_width = 20

    def get_value(self):
        return self.data['error']

COLUMNS['error'] = Error


class ErrorAnnounce(ColumnBase):
    header = {'left': 'Announce Error'}
    align = 'left'
    width = None
    min_width = 10

    def get_value(self):
        return self.data['error-announce']

COLUMNS['error-announce'] = ErrorAnnounce


class ErrorScrape(ColumnBase):
    header = {'left': 'Scrape Error'}
    align = 'left'
    width = None
    min_width = 10

    def get_value(self):
        return self.data['error-scrape']

COLUMNS['error-scrape'] = ErrorScrape


class Downloads(ColumnBase):
    header = {'left': 'Downloads'}
    align = 'right'
    width = 9
    min_width = 5

    def get_value(self):
        return self.data['count-downloads']

COLUMNS['downloads'] = Downloads


class Leeches(ColumnBase):
    header = {'left': 'Leeches'}
    align = 'right'
    width = 7
    min_width = 5

    def get_value(self):
        return self.data['count-leeches']

COLUMNS['leeches'] = Leeches


class Seeds(ColumnBase):
    header = {'left': 'Seeds'}
    align = 'right'
    width = 5
    min_width = 5

    def get_value(self):
        return self.data['count-seeds']

COLUMNS['seeds'] = Seeds


class LastAnnounce(ColumnBase):
    header = {'left': 'Last Announce'}
    align = 'right'
    width = 13
    min_width = 10

    def get_value(self):
        return self.data['time-last-announce']

COLUMNS['last-announce'] = LastAnnounce


class NextAnnounce(ColumnBase):
    header = {'left': 'Next Announce'}
    align = 'right'
    width = 13
    min_width = 10

    def get_value(self):
        return self.data['time-next-announce']

COLUMNS['next-announce'] = NextAnnounce


class LastScrape(ColumnBase):
    header = {'left': 'Last Scrape'}
    align = 'right'
    width = 11
    min_width = 10

    def get_value(self):
        return self.data['time-last-scrape']

COLUMNS['last-scrape'] = LastScrape


class NextScrape(ColumnBase):
    header = {'left': 'Next Scrape'}
    align = 'right'
    width = 11
    min_width = 10

    def get_value(self):
        return self.data['time-next-scrape']

COLUMNS['next-scrape'] = NextScrape



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...client import PeerFilter
from .base import ItemWidgetBase, ListWidgetBase
from .peer import TUICOLUMNS
from .utils import stringify_torrent_filter


class PeerItemWidget(ItemWidgetBase):
    palette_unfocused = 'peerlist'

    @property
    def id(self):
        return self.data['id']

    @property
    def torrent_id(self):
        return self.data['tid']


class PeerListWidget(ListWidgetBase):
    tuicolumns      = TUICOLUMNS
    ListItemClass   = PeerItemWidget
    keymap_context  = 'peer'
    palette_name    = 'peerlist'
    focusable_items = False

    def __init__(self, srvapi, keymap, tfilter=None, pfilter=None, columns=None, sort=None, title=None):
        super().__init__(srvapi, keymap, columns=columns, sort=sort, title=title)
        self._tfilter = tfilter
        self._pfilter = pfilter
        self._secondary_filter = None

        # Create peer filter generator
        if pfilter is not None:
            def filter_peers(peers):
                yield from pfilter.apply(peers)
        else:
            def filter_peers(peers):
                yield from peers
        self._maybe_filter_peers = filter_peers

        self._poller = self._srvapi.create_poller(
            self._srvapi.torrent.torrents, tfilter, keys=('peers', 'name', 'id')
        )
        self._poller.on_response(self._handle_peers)

    def _handle_peers(self, response):
        if response is None or not response.torrents:
            self.clear()
        else:
            # Auto-generate title from our filters if not set
            if self._title_name is None:
                self._title_name = stringify_torrent_filter(self._tfilter, response.torrents)
                if self._pfilter:
                    self._title_name += ' %s' % self._pfilter

            # Create list items our base widget can handle
            def peers_combined(torrents):
                for t in torrents:
                    yield from self._maybe_filter_peers(t['peers'])
            self._data_dict = {p['id']:p for p in peers_combined(response.torrents)}
        self._invalidate()

    def clear(self):
        for w in self._listbox.body:
            w.data.clearcache()
        super().clear()

    def refresh(self):
        self._poller.poll()

    @property
    def sort(self):
        return self._sort

    @sort.setter
    def sort(self, sort):
        ListWidgetBase.sort.fset(self, sort)
        self._poller.poll()

    @property
    def secondary_filter(self):
        return self._secondary_filter

    @secondary_filter.setter
    def secondary_filter(self, peer_filter):
        if peer_filter is None:
            self._secondary_filter = None
        else:
            self._secondary_filter = PeerFilter(peer_filter)
        self._invalidate()

    def _limit_items(self, peer_widgets):
        # Combine primary and secondary peer filters
        pfilter = self._pfilter
        spfilter = self._secondary_filter
        if pfilter is None:
            pfilter = spfilter
        elif spfilter is not None:
            pfilter = pfilter & spfilter

        if pfilter is not None:
            for pw in peer_widgets:
                if not pfilter.match(pw.data):
                    yield pw



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Mixin classes that are common between TUI and CLI commands"""

import os

from ... import objects
from ...settings import defaults


class get_single_torrent():
    async def get_single_torrent(self, tfilter, keys=(), one_or_none=False):
        """
        Get a single torrent that matches TorrentFilter `tfilter`

        A single Torrent instance is returned if there is only one match or if
        one_or_none evaluates to False, in which case all torrents are sorted by
        name and the first one is returned.

        Return None if no match is found or if there are multiple matches and
        `one_or_none` evaluates to True.
        """
        keys = tuple(keys)
        if 'name' not in keys:
            keys = keys + ('name',)
        request = objects.srvapi.torrent.torrents(tfilter, keys=keys)
        response = await self.make_request(request, polling_frenzy=False, quiet=True)
        if response.success:
            if len(response.torrents) == 1 or not one_or_none:
                from ...client import TorrentSorter
                torrents = TorrentSorter(('name',)).apply(response.torrents)
                return torrents[0]

class get_torrent_sorter():
    def get_torrent_sorter(self, args):
        """
        Return TorrentSorter instance or None

        If `args` evaluates to True, it is passed to TorrentSorter and the
        result is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import TorrentSorter
            return TorrentSorter(objects.localcfg.validate('sort.torrents', args))

class get_torrent_columns():
    def get_torrent_columns(self, columns):
        """
        Check if each item in iterable `columns` is a valid torrent list column name

        Raise ValueError or return a new list of `columns`.
        """
        return objects.localcfg.validate('columns.torrents', columns)



class get_file_columns():
    def get_file_columns(self, columns):
        """
        Check if each item in iterable `columns` is a valid file list column name

        Raise ValueError or return a new list of `columns`.
        """
        return objects.localcfg.validate('columns.files', columns)



class get_peer_sorter():
    def get_peer_sorter(self, args):
        """
        Return PeerSorter instance or None

        If `args` evaluates to True, it is passed to PeerSorter and the result
        is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import PeerSorter
            return PeerSorter(objects.localcfg.validate('sort.peers', args))

class get_peer_filter():
    def get_peer_filter(self, FILTER):
        """
        Return PeerFilter instance or None

        If `FILTER` evaluates to True, it is passed to PeerFilter and the
        resulting object is returned.

        If `FILTER` evaluates to False, None is returned.
        """
        if FILTER:
            from ...client import PeerFilter
            return PeerFilter(FILTER)

class get_peer_columns():
    def get_peer_columns(self, columns):
        """
        Check if each item in iterable `columns` is a valid peer list column name

        Raise ValueError or return a new list of `columns`.
        """
        return objects.localcfg.validate('columns.peers', columns)



class get_tracker_sorter():
    def get_tracker_sorter(self, args):
        """
        Return TrackerSorter instance or None

        If `args` evaluates to True, it is passed to TrackerSorter and
        the result is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import TrackerSorter
            return TrackerSorter(objects.localcfg.validate('sort.trackers', args))

class get_tracker_filter():
    def get_tracker_filter(self, FILTER):
        """
        Return TrackerFilter instance or None

        If `FILTER` evaluates to True, it is passed to TrackerFilter and
        the resulting object is returned.

        If `FILTER` evaluates to False, None is returned.
        """
        if FILTER:
            from ...client import TrackerFilter
            return TrackerFilter(FILTER)

class get_tracker_columns():
    def get_tracker_columns(self, columns):
        """
        Check if each item in iterable `columns` is a valid tracker list column name

        Raise ValueError or return a new list of `columns`.
        """
        return objects.localcfg.validate('columns.trackers', columns)



class get_setting_sorter():
    def get_setting_sorter(self, args):
        """
        Return SettingSorter instance or None

        If `args` evaluates to True, it is passed to SettingSorter and the
        result is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import SettingSorter
            return SettingSorter(objects.localcfg.validate('sort.settings', args))

class get_setting_columns():
    def get_setting_columns(self, columns):
        """
        Check if each item in iterable `columns` is a valid setting list column name

        Raise ValueError or return a new list of `columns`.
        """
        return objects.localcfg.validate('columns.settings', columns)


class get_rc_filepath():
    def get_rc_filepath(self, path):
        """Return `path` relative to default rc file path unless it is absolute."""
        path = os.path.expanduser(path)
        if not os.path.exists(path) and \
           not os.path.isabs(path) and \
           not path.startswith('.{sep}'.format(sep=os.sep)) and \
           not path.startswith('~'):
            default_dir = os.path.dirname(defaults.DEFAULT_RCFILE)
            path = os.path.join(default_dir, path)
        return path
